"""
Receiver Gather (version 1.6-1) bindings to ObsPy core module.
"""
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)
from future.builtins import *  # NOQA

from collections import namedtuple

import numpy as np

from obspy.core import Stream, Trace, Stats, UTCDateTime
from obspy.io.rg16.util import _read, _open_file, _quick_merge


HeaderCount = namedtuple('HeaderCount', 'channel_set extended external')


@_open_file
def _read_rg16(filename, headonly=False, starttime=None, endtime=None,
               merge=False, contacts_north=False, details=False, **kwargs):
    """
    Read Fairfield Nodal's Receiver Gather File Format version 1.6-1.

    .. warning::
        This function should NOT be called directly, it registers via the
        ObsPy :func:`~obspy.core.stream.read` function, call this instead.

    :param filename: path to the rg16 file or a file object.
    :type filename: str, buffer
    :param headonly: If True don't read data, only main information
        contained in the headers of the trace block is read.
    :type headonly: optional, bool
    :param starttime: If not None dont read traces that start before starttime.
    :type starttime: optional, :class:`~obspy.core.utcdatetime.UTCDateTime`
    :param endtime: If not None dont read traces that start after endtime.
    :type endtime: optional, :class:`~obspy.core.utcdatetime.UTCDateTime`
    :param merge: If True merge contiguous data blocks as they are found. For
        continuous data files having 100,000+ traces this will create
        more manageable streams.
    :type merge: bool
    :param contacts_north: If this parameter is set to True, it will map the
        components to Z (1C, 3C), N (3C), and E (3C) as well as correct
        the polarity for the vertical component.
    :type contacts_north: bool
    :param details: If True, all the information contained in the headers
        is read).
    :type details: bool
    :return: An ObsPy :class:`~obspy.core.stream.Stream` object.
        Frequencies are expressed in hertz and time is expressed in second
        (except for date).
    """
    starttime = starttime or UTCDateTime(1970, 1, 1)
    endtime = endtime or UTCDateTime()
    # get the number of headers/records, position of trace data
    # and record length.
    header_count = _cmp_nbr_headers(filename)
    record_count = _cmp_nbr_records(filename)
    trace_block_start = 32 * (2 + sum(header_count))
    record_length = _cmp_record_length(filename)
    # create trace data
    traces = []
    for i in range(0, record_count):
        nbr_bytes_trace_block = _cmp_jump(filename, trace_block_start)
        trace_starttime = _read(filename, trace_block_start + 20 + 2 * 32, 8,
                                'binary') / 1e6
        trace_endtime = trace_starttime + record_length
        con1 = starttime.timestamp > trace_endtime
        con2 = endtime.timestamp < trace_starttime
        # determine if this block is excluded based in starttime/endtime params
        if con1 or con2:
            # the block has to be excluded, increment trace_block_start
            #  and continue
            trace_block_start += nbr_bytes_trace_block
            continue
        trace = _make_trace(filename, trace_block_start, headonly,
                            contacts_north, details)
        traces.append(trace)
        trace_block_start += nbr_bytes_trace_block
    if merge:
        traces = _quick_merge(traces)
    return Stream(traces=traces)


def _cmp_nbr_headers(fi):
    """
    Return a tuple containing the number of channel set headers,
    the number of extended headers and the number of external headers
    in the file.
    """
    header_count = HeaderCount(
        channel_set=_read(fi, 28, 1, 'bcd'),
        extended=_read(fi, 37, 2, 'binary'),
        external=_read(fi, 39, 3, 'binary'),
    )
    return header_count


def _cmp_nbr_records(fi):
    """
    Return the number of records in the file (ie number of time slices
    multiplied by the number of components).
    """
    initial_header = _read_initial_headers(fi)
    channel_sets_descriptor = initial_header['channel_sets_descriptor']
    channels_number = set()

    for _, val in channel_sets_descriptor.items():
        channels_number.add(val['RU_channel_number'])
    nbr_component = len(channels_number)
    extended_header_2 = initial_header['extended_headers']['2']
    nbr_time_slices = extended_header_2['nbr_time_slices']
    nbr_records = nbr_time_slices * nbr_component
    return nbr_records


def _cmp_record_length(fi):
    """
    Return the record length.
    """
    base_scan_interval = _read(fi, 22, 1, 'binary')
    sampling_rate = int(1000 / (base_scan_interval / 16))
    gen_head_2 = _read_initial_headers(fi)['general_header_2']
    record_length = gen_head_2['extended_record_length'] - 1 / sampling_rate
    return record_length


def _cmp_jump(fi, trace_block_start):
    """
    Return the number of bytes in a trace block.
    """
    nbr_trace_extension_block = _read(fi, trace_block_start + 9, 1, 'binary')
    nbr_bytes_header_trace = 20 + 32 * nbr_trace_extension_block
    nbr_sample_trace = _read(fi, trace_block_start + 27, 3, 'binary')
    nbr_bytes_trace_data = nbr_sample_trace * 4
    nbr_bytes_trace_block = nbr_bytes_trace_data + nbr_bytes_header_trace
    return nbr_bytes_trace_block


def _make_trace(fi, trace_block_start, headonly, standard_orientation,
                details):
    """
    Make obspy trace from a trace block (header + trace).
    """
    stats = _make_stats(fi, trace_block_start, standard_orientation, details)
    if headonly:
        data = np.array([])
    else:  # read trace
        nbr_trace_extension_block = _read(fi, trace_block_start + 9,
                                          1, 'binary')
        trace_start = trace_block_start + 20 + nbr_trace_extension_block * 32
        nbr_sample_trace = _read(fi, trace_block_start + 27, 3, 'binary')
        nbr_bytes_trace = 4 * nbr_sample_trace
        data = _read(fi, trace_start, nbr_bytes_trace, 'IEEE')
        if stats.channel[-1] == 'Z':
            data = -data
            data = data.astype('>f4')
    return Trace(data=data, header=stats)


def _make_stats(fi, tr_block_start, standard_orientation, details):
    """
    Make Stats object from information contained in the header of the trace.
    """
    base_scan_interval = _read(fi, 22, 1, 'binary')
    sampling_rate = int(1000 / (base_scan_interval / 16))
    # map sampling rate to band code according to seed standard
    band_map = {2000: 'G', 1000: 'G', 500: 'D', 250: 'D'}
    # geophone instrument code
    instrument_code = 'P'
    # mapping for "standard_orientation"
    standard_component_map = {'2': 'Z', '3': 'N', '4': 'E'}
    component = str(_read(fi, tr_block_start + 40, 1, 'binary'))
    if standard_orientation:
        component = standard_component_map[component]
    chan = band_map[sampling_rate] + instrument_code + component
    npts = _read(fi, tr_block_start + 27, 3, 'binary')
    start_time = _read(fi, tr_block_start + 20 + 2 * 32, 8, 'binary') / 1e6
    end_time = start_time + (npts - 1) * (1 / sampling_rate)
    network = _read(fi, tr_block_start + 20, 3, 'binary')
    station = _read(fi, tr_block_start + 23, 3, 'binary')
    location = _read(fi, tr_block_start + 26, 1, 'binary')
    statsdict = dict(starttime=UTCDateTime(start_time),
                     endtime=UTCDateTime(end_time),
                     sampling_rate=sampling_rate,
                     npts=npts,
                     network=str(network),
                     station=str(station),
                     location=str(location),
                     channel=chan)
    if details:
        statsdict['rg16'] = {}
        statsdict['rg16']['initial_headers'] = {}
        stats_initial_headers = statsdict['rg16']['initial_headers']
        stats_initial_headers.update(_read_initial_headers(fi))
        statsdict['rg16']['trace_headers'] = {}
        stats_tr_headers = statsdict['rg16']['trace_headers']
        stats_tr_headers.update(_read_trace_header(fi, tr_block_start))
        nbr_tr_header_block = _read(fi, tr_block_start + 9,
                                    1, 'binary')
        if nbr_tr_header_block > 0:
            stats_tr_headers.update(
                _read_trace_headers(fi, tr_block_start, nbr_tr_header_block))
    return Stats(statsdict)


def _read_trace_headers(fi, trace_block_start, nbr_trace_header):
    """
    Read headers in the trace block.
    """
    trace_headers = {}
    dict_func = {'1': _read_trace_header_1, '2': _read_trace_header_2,
                 '3': _read_trace_header_3, '4': _read_trace_header_4,
                 '5': _read_trace_header_5, '6': _read_trace_header_6,
                 '7': _read_trace_header_7, '8': _read_trace_header_8,
                 '9': _read_trace_header_9, '10': _read_trace_header_10}
    for i in range(1, nbr_trace_header + 1):
        trace_headers.update(dict_func[str(i)](fi, trace_block_start))
    return trace_headers


def _read_trace_header(fi, trace_block_start):
    """
    Read the 20 bytes trace header (first header in the trace block).
    """
    trace_number = _read(fi, trace_block_start + 4, 2, 'bcd')
    trace_edit_code = _read(fi, trace_block_start + 11, 1, 'binary')
    return {'trace_number': trace_number, 'trace_edit_code': trace_edit_code}


def _read_trace_header_1(fi, trace_block_start):
    """
    Read trace header 1
    """
    pos = trace_block_start + 20

    dict_header_1 = dict(
        extended_receiver_line_nbr=_read(fi, pos + 10, 5, 'binary'),
        extended_receiver_point_nbr=_read(fi, pos + 15, 5, 'binary'),
        sensor_type=_read(fi, pos + 20, 1, 'binary'),
        trace_count_file=_read(fi, pos + 21, 4, 'binary'),
    )
    return dict_header_1


def _read_trace_header_2(fi, trace_block_start):
    """
    Read trace header 2
    """
    pos = trace_block_start + 20 + 32

    leg_source_info = {'0': 'undefined', '1': 'preplan', '2': 'as shot',
                       '3': 'post processed'}
    source_key = str(_read(fi, pos + 29, 1, 'binary'))

    leg_energy_source = {'0': 'undefined', '1': 'vibroseis', '2': 'dynamite',
                         '3': 'air gun'}
    energy_source_key = str(_read(fi, pos + 30, 1, 'binary'))

    dict_header_2 = dict(
        shot_line_nbr=_read(fi, pos, 4, 'binary'),
        shot_point=_read(fi, pos + 4, 4, 'binary'),
        shot_point_index=_read(fi, pos + 8, 1, 'binary'),
        shot_point_pre_plan_x=_read(fi, pos + 9, 4, 'binary') / 10,
        shot_point_pre_plan_y=_read(fi, pos + 13, 4, 'binary') / 10,
        shot_point_final_x=_read(fi, pos + 17, 4, 'binary') / 10,
        shot_point_final_y=_read(fi, pos + 21, 4, 'binary') / 10,
        shot_point_final_depth=_read(fi, pos + 25, 4, 'binary') / 10,
        source_of_final_shot_info=leg_source_info[source_key],
        energy_source_type=leg_energy_source[energy_source_key],
    )
    return dict_header_2


def _read_trace_header_3(fi, trace_block_start):
    """
    Read trace header 3
    """
    pos = trace_block_start + 20 + 32 * 2

    dict_header_3 = dict(
        epoch_time=UTCDateTime(_read(fi, pos, 8, 'binary') / 1e6),
        # shot skew time in second
        shot_skew_time=_read(fi, pos + 8, 8, 'binary') / 1e6,
        # time shift clock correction in second
        time_shift_clock_correction=_read(fi, pos + 16, 8, 'binary') / 1e9,
        # remaining clock correction in second
        remaining_clock_correction=_read(fi, pos + 24, 8, 'binary') / 1e9,
    )
    return dict_header_3


def _read_trace_header_4(fi, trace_block_start):
    """
    Read trace header 4
    """
    pos = trace_block_start + 20 + 32 * 3

    leg_trace_clipped = {'0': 'not clipped', '1': 'digital clip detected',
                         '2': 'analog clip detected'}
    clipped_code = str(_read(fi, pos + 9, 1, 'binary'))

    leg_record_type = {'2': 'test data record',
                       '8': 'normal seismic data record'}
    record_type_code = str(_read(fi, pos + 10, 1, 'binary'))

    leg_shot_flag = {'0': 'normal', '1': 'bad-operator specified',
                     '2': 'bad-failed to QC test'}
    shot_code = str(_read(fi, pos + 11, 1, 'binary'))

    dict_header_4 = dict(
        # pre shot guard band in second
        pre_shot_guard_band=_read(fi, pos, 4, 'binary') / 1e3,
        # post shot guard band in second
        post_shot_guard_band=_read(fi, pos + 4, 4, 'binary') / 1e3,
        # preamp gain in dB
        preamp_gain=_read(fi, pos + 8, 1, 'binary'),
        trace_clipped_flag=leg_trace_clipped[clipped_code],
        record_type_code=leg_record_type[record_type_code],
        shot_status_flag=leg_shot_flag[shot_code],
        external_shot_id=_read(fi, pos + 12, 4, 'binary'),
        post_processed_first_break_pick_time=_read(fi, pos + 24, 4, 'IEEE'),
        post_processed_rms_noise=_read(fi, pos + 28, 4, 'IEEE'),
    )
    return dict_header_4


def _read_trace_header_5(fi, trace_block_start):
    """
    Read trace header 5
    """
    pos = trace_block_start + 20 + 32 * 4

    leg_source_receiver_info = {
        '1': 'preplan',
        '2': 'as laid (no navigation sensor)',
        '3': 'as laid (HiPAP only)',
        '4': 'as laid (HiPAP and INS)',
        '5': 'as laid (HiPAP and DVL)',
        '6': 'as laid (HiPAP, DVL and INS)',
        '7': 'post processed (HiPAP only)',
        '8': 'post processed (HiPAP and INS)',
        '9': 'post processed (HiPAP and DVL)',
        '10': 'post processed (HiPAP, DVL ans INS)',
        '11': 'first break analysis',
    }
    source_key = str(_read(fi, pos + 29, 1, 'binary'))

    dict_header_5 = dict(
        receiver_point_pre_plan_x=_read(fi, pos + 9, 4, 'binary') / 10,
        receiver_point_pre_plan_y=_read(fi, pos + 13, 4, 'binary') / 10,
        receiver_point_final_x=_read(fi, pos + 17, 4, 'binary') / 10,
        receiver_point_final_y=_read(fi, pos + 21, 4, 'binary') / 10,
        receiver_point_final_depth=_read(fi, pos + 25, 4, 'binary') / 10,
        source_of_final_receiver_info=leg_source_receiver_info[source_key],
    )
    return dict_header_5


def _read_trace_header_6(fi, trace_block_start):
    """
    Read trace header 6
    """
    pos = trace_block_start + 20 + 32 * 5

    dict_header_6 = dict(
        tilt_matrix_h1x=_read(fi, pos, 4, 'IEEE'),
        tilt_matrix_h2x=_read(fi, pos + 4, 4, 'IEEE'),
        tilt_matrix_vx=_read(fi, pos + 8, 4, 'IEEE'),
        tilt_matrix_h1y=_read(fi, pos + 12, 4, 'IEEE'),
        tilt_matrix_h2y=_read(fi, pos + 16, 4, 'IEEE'),
        tilt_matrix_vy=_read(fi, pos + 20, 4, 'IEEE'),
        tilt_matrix_h1z=_read(fi, pos + 24, 4, 'IEEE'),
        tilt_matrix_h2z=_read(fi, pos + 28, 4, 'IEEE'),
    )
    return dict_header_6


def _read_trace_header_7(fi, trace_block_start):
    """
    Read trace header 7
    """
    pos = trace_block_start + 20 + 32 * 6

    dict_header_7 = dict(
        tilt_matrix_vz=_read(fi, pos, 4, 'IEEE'),
        azimuth_degree=_read(fi, pos + 4, 4, 'IEEE'),
        pitch_degree=_read(fi, pos + 8, 4, 'IEEE'),
        roll_degree=_read(fi, pos + 12, 4, 'IEEE'),
        remote_unit_temp=_read(fi, pos + 16, 4, 'IEEE'),
        remote_unit_humidity=_read(fi, pos + 20, 4, 'IEEE'),
        orientation_matrix_version_nbr=_read(fi, pos + 24, 4, 'binary'),
        gimbal_corrections=_read(fi, pos + 28, 1, 'binary'))
    return dict_header_7


def _read_trace_header_8(fi, trace_block_start):
    """
    Read trace header 8
    """
    pos = trace_block_start + 20 + 32 * 7

    leg_preamp_path = {
        '0': 'external input selected',
        '1': 'simulated data selected',
        '2': 'pre-amp input shorted to ground',
        '3': 'test oscillator with sensors',
        '4': 'test oscillator without sensors',
        '5': 'common mode test oscillator with sensors',
        '6': 'common mode test oscillator without sensors',
        '7': 'test oscillator on positive sensors with neg sensor grounded',
        '8': 'test oscillator on negative sensors with pos sensor grounded',
        '9': 'test oscillator on positive PA input with neg PA input ground',
        '10': 'test oscillator on negative PA input with pos PA input ground',
        '11': 'test oscillator on positive PA input with neg\
                              PA input ground, no sensors',
        '12': 'test oscillator on negative PA input with pos\
                              PA input ground, no sensors'}
    preamp_path_code = str(_read(fi, pos + 24, 4, 'binary'))

    leg_test_oscillator = {'0': 'test oscillator path open',
                           '1': 'test signal selected',
                           '2': 'DC reference selected',
                           '3': 'test oscillator path grounded',
                           '4': 'DC reference toggle selected'}
    oscillator_code = str(_read(fi, pos + 28, 4, 'binary'))

    dict_header_8 = dict(
        fairfield_test_analysis_code=_read(fi, pos, 4, 'binary'),
        first_test_oscillator_attenuation=_read(fi, pos + 4, 4, 'binary'),
        second_test_oscillator_attenuation=_read(fi, pos + 8, 4, 'binary'),
        # start delay in second
        start_delay=_read(fi, pos + 12, 4, 'binary') / 1e6,
        dc_filter_flag=_read(fi, pos + 16, 4, 'binary'),
        dc_filter_frequency=_read(fi, pos + 20, 4, 'IEEE'),
        preamp_path=leg_preamp_path[preamp_path_code],
        test_oscillator_signal_type=leg_test_oscillator[oscillator_code],
    )
    return dict_header_8


def _read_trace_header_9(fi, trace_block_start):
    """
    Read trace header 9
    """
    pos = trace_block_start + 20 + 32 * 8

    leg_signal_type = {'0': 'pattern is address ramp',
                       '1': 'pattern is RU address ramp',
                       '2': 'pattern is built from provided values',
                       '3': 'pattern is random numbers',
                       '4': 'pattern is a walking 1s',
                       '5': 'pattern is a walking 0s',
                       '6': 'test signal is a specified DC value',
                       '7': 'test signal is a pulse train with\
                             specified duty cycle',
                       '8': 'test signal is a sine wave',
                       '9': 'test signal is a dual tone sine',
                       '10': 'test signal is an impulse',
                       '11': 'test signal is a step function'}
    type_code = str(_read(fi, pos, 4, 'binary'))

    # test signal generator frequency 1 in hertz
    test_signal_freq_1 = _read(fi, pos + 4, 4, 'binary') / 1e3
    # test signal generator frequency 2 in hertz
    test_signal_freq_2 = _read(fi, pos + 8, 4, 'binary') / 1e3
    # test signal generator amplitude 1 in dB down from full scale -120 to 120
    test_signal_amp_1 = _read(fi, pos + 12, 4, 'binary')
    # test signal generator amplitude 2 in dB down from full scale -120 to 120
    test_signal_amp_2 = _read(fi, pos + 16, 4, 'binary')
    # test signal generator duty cycle in percentage
    duty_cycle = _read(fi, pos + 20, 4, 'IEEE')
    # test signal generator active duration in second
    active_duration = _read(fi, pos + 24, 4, 'binary') / 1e6
    # test signal generator activation time in second
    activation_time = _read(fi, pos + 28, 4, 'binary') / 1e6

    dict_header_9 = dict(
        test_signal_generator_signal_type=leg_signal_type[type_code],
        test_signal_generator_frequency_1=test_signal_freq_1,
        test_signal_generator_frequency_2=test_signal_freq_2,
        test_signal_generator_amplitude_1=test_signal_amp_1,
        test_signal_generator_amplitude_2=test_signal_amp_2,
        test_signal_generator_duty_cycle_percentage=duty_cycle,
        test_signal_generator_active_duration=active_duration,
        test_signal_generator_activation_time=activation_time,
    )
    return dict_header_9


def _read_trace_header_10(fi, trace_block_start):
    """
    Read trace header 10
    """
    pos = trace_block_start + 20 + 32 * 9

    dict_header_10 = dict(
        test_signal_generator_idle_level=_read(fi, pos, 4, 'binary'),
        test_signal_generator_active_level=_read(fi, pos + 4, 4, 'binary'),
        test_signal_generator_pattern_1=_read(fi, pos + 8, 4, 'binary'),
        test_signal_generator_pattern_2=_read(fi, pos + 12, 4, 'binary'),
    )
    return dict_header_10


@_open_file
def _is_rg16(filename, **kwargs):
    """
    Determine if a file is a rg16 file.

    :param filename: a path to a file or a file object
    :type filename: str, buffer
    :rtype: bool
    :return: True if the file object is a rg16 file.
    """
    try:
        sample_format = _read(filename, 2, 2, 'bcd')
        manufacturer_code = _read(filename, 16, 1, 'bcd')
        version = _read(filename, 42, 2, 'binary')
    except ValueError:  # if file too small
        return False
    con1 = version == 262 and sample_format == 8058
    return con1 and manufacturer_code == 20


@_open_file
def _read_initial_headers(filename):
    """
    Extract all the information contained in the headers located before data,
    at the beginning of the rg16 file object.

    :param filename : a path to a rg16 file or a rg16 file object.
    :type filename: str, buffer
    :return: a dictionnary containing all the information of the initial
        headers

    Frequencies are expressed in hertz and time is expressed in second
    (except for the date).
    """
    headers_content = dict(
        general_header_1=_read_general_header_1(filename),
        general_header_2=_read_general_header_2(filename),
        channel_sets_descriptor=_read_channel_sets(filename),
        extended_headers=_read_extended_headers(filename),
    )
    return headers_content


def _read_general_header_1(fi):
    """
    Extract information contained in the general header block 1
    """
    gen_head_1 = dict(
        file_number=_read(fi, 0, 2, 'bcd'),
        sample_format_code=_read(fi, 2, 2, 'bcd'),
        general_constant=_read(fi, 4, 6, 'bcd'),
        time_slice_year=_read(fi, 10, 1, 'bcd'),
        nbr_add_general_header=_read(fi, 11, 0.5, 'bcd'),
        julian_day=_read(fi, 11, 1.5, 'bcd', False),
        time_slice=_read(fi, 13, 3, 'bcd'),
        manufacturer_code=_read(fi, 16, 1, 'bcd'),
        manufacturer_serial_number=_read(fi, 17, 2, 'bcd'),
        base_scan_interval=_read(fi, 22, 1, 'binary'),
        polarity_code=_read(fi, 23, 0.5, 'binary'),
        record_type=_read(fi, 25, 0.5, 'binary'),
        scan_type_per_record=_read(fi, 27, 1, 'bcd'),
        nbr_channel_set=_read(fi, 28, 1, 'bcd'),
        nbr_skew_block=_read(fi, 29, 1, 'bcd'),
    )
    return gen_head_1


def _read_general_header_2(fi):
    """
    Extract information contained in the general header block 2
    """
    gen_head_2 = dict(
        extended_file_number=_read(fi, 32, 3, 'binary'),
        extended_channel_sets_per_scan_type=_read(fi, 35, 2, 'binary'),
        extended_header_blocks=_read(fi, 37, 2, 'binary'),
        external_header_blocks=_read(fi, 39, 3, 'binary'),
        version_number=_read(fi, 42, 2, 'binary'),
        # extended record length in second
        extended_record_length=_read(fi, 46, 3, 'binary') / 1e3,
        general_header_block_number=_read(fi, 50, 1, 'binary'),
    )
    return gen_head_2


def _read_channel_sets(fi):
    """
    Extract information of all channel set descriptor blocks.
    """
    channel_sets = {}
    nbr_channel_set = _read(fi, 28, 1, 'bcd')
    start_byte = 64
    for i in range(0, nbr_channel_set):
        channel_set_name = str(i + 1)
        channel_sets[channel_set_name] = _read_channel_set(fi, start_byte)
        start_byte += 32
    return channel_sets


def _read_channel_set(fi, start_byte):
    """
    Extract information contained in the ith channel set descriptor.
    """
    nbr_32_ext = _read(fi, start_byte + 28, 0.5, 'binary', False)

    channel_set = dict(
        scan_type_number=_read(fi, start_byte, 1, 'bcd'),
        channel_set_number=_read(fi, start_byte + 1, 1, 'bcd'),
        channel_set_start_time=_read(fi, start_byte + 2, 2, 'binary') * 2e-3,
        channel_set_end_time=_read(fi, start_byte + 4, 2, 'binary') * 2e-3,
        optionnal_MP_factor=_read(fi, start_byte + 6, 1, 'binary'),
        mp_factor_descaler_multiplier=_read(fi, start_byte + 7, 1, 'binary'),
        nbr_channels_in_channel_set=_read(fi, start_byte + 8, 2, 'bcd'),
        channel_type_code=_read(fi, start_byte + 10, 0.5, 'binary'),
        nbr_sub_scans=_read(fi, start_byte + 11, 0.5, 'bcd'),
        gain_control_type=_read(fi, start_byte + 11, 0.5, 'bcd', False),
        alias_filter_frequency=_read(fi, start_byte + 12, 2, 'bcd'),
        alias_filter_slope=_read(fi, start_byte + 14, 2, 'bcd'),
        low_cut_filter_freq=_read(fi, start_byte + 16, 2, 'bcd'),
        low_cut_filter_slope=_read(fi, start_byte + 18, 2, 'bcd'),
        notch_filter_freq=_read(fi, start_byte + 20, 2, 'bcd') / 10,
        notch_2_filter_freq=_read(fi, start_byte + 22, 2, 'bcd') / 10,
        notch_3_filter_freq=_read(fi, start_byte + 24, 2, 'bcd') / 10,
        extended_channel_set_number=_read(fi, start_byte + 26, 2, 'binary'),
        extended_header_flag=_read(fi, start_byte + 28, 0.5, 'binary'),
        nbr_32_byte_trace_header_extension=nbr_32_ext,
        vertical_stack_size=_read(fi, start_byte + 29, 1, 'binary'),
        RU_channel_number=_read(fi, start_byte + 30, 1, 'binary'),
        array_forming=_read(fi, start_byte + 31, 1, 'binary'),

    )

    return channel_set


def _read_extended_headers(fi):
    """
    Extract information from the extended headers.
    """
    extended_headers = {}
    nbr_channel_set = _read(fi, 28, 1, 'bcd')
    start_byte = 32 + 32 + 32 * nbr_channel_set
    extended_headers['1'] = _read_extended_header_1(fi, start_byte)
    start_byte += 32
    extended_headers['2'] = _read_extended_header_2(fi, start_byte)
    start_byte += 32
    extended_headers['3'] = _read_extended_header_3(fi, start_byte)
    nbr_extended_headers = _read(fi, 37, 2, 'binary', True)
    if nbr_extended_headers > 3:
        coeffs = extended_headers['2']['number_decimation_filter_coefficient']
        nbr_coeff_remain = coeffs % 8
        for i in range(3, nbr_extended_headers):
            start_byte += 32
            extended_header_name = str(i + 1)
            if i == nbr_extended_headers - 1:
                header = _read_extended_header(fi, start_byte, i + 1,
                                               nbr_coeff_remain)
                extended_headers[extended_header_name] = header
            else:
                header = _read_extended_header(fi, start_byte, i + 1, 8)
                extended_headers[extended_header_name] = header
    return extended_headers


def _read_extended_header_1(fi, start_byte):
    """
    Extract information contained in the extended header block number 1.
    """
    deployment_time = _read(fi, start_byte + 8, 8, 'binary') / 1e6
    pick_up_time = _read(fi, start_byte + 16, 8, 'binary') / 1e6
    start_time_ru = _read(fi, start_byte + 24, 8, 'binary') / 1e6

    extended_header_1 = dict(
        id_ru=_read(fi, start_byte, 8, 'binary'),
        deployment_time=UTCDateTime(deployment_time),
        pick_up_time=UTCDateTime(pick_up_time),
        start_time_ru=UTCDateTime(start_time_ru),
    )
    return extended_header_1


def _read_extended_header_2(fi, start_byte):
    """
    Extract information contained in the extended header block number 2.
    """
    # code mappings to meaning
    leg_clock_stop = {'0': 'normal', '1': 'storage full', '2': 'power loss',
                      '3': 'reboot'}
    stop_code = str(_read(fi, start_byte + 12, 1, 'binary'))

    leg_freq_drift = {'0': 'not within specification',
                      '1': 'within specification'}
    drift_code = str(_read(fi, start_byte + 13, 1, 'binary'))

    leg_oscillator_type = {'0': 'control board', '1': 'atomic',
                           '2': 'ovenized', '3': 'double ovenized',
                           '4': 'disciplined'}
    oscillator_code = str(_read(fi, start_byte + 14, 1, 'binary'))

    leg_data_collection = {'0': 'normal', '1': 'continuous',
                           '2': 'shot sliced with guard band'}
    data_collection_code = str(_read(fi, start_byte + 15, 1, 'binary'))

    leg_data_decimation = {'0': 'not decimated', '1': 'decimated data'}
    decimation_code = str(_read(fi, start_byte + 28, 1, 'binary'))

    extended_header_2 = dict(
        acquisition_drift_window=_read(fi, start_byte, 4, 'IEEE') * 1e-6,
        clock_drift=_read(fi, start_byte + 4, 8, 'binary') * 1e-9,
        clock_stop_method=leg_clock_stop[stop_code],
        frequency_drift=leg_freq_drift[drift_code],
        oscillator_type=leg_oscillator_type[oscillator_code],
        data_collection_method=leg_data_collection[data_collection_code],
        nbr_time_slices=_read(fi, start_byte + 16, 4, 'binary'),
        nbr_files=_read(fi, start_byte + 20, 4, 'binary'),
        file_number=_read(fi, start_byte + 24, 4, 'binary'),
        data_decimation=leg_data_decimation[decimation_code],
        original_base_scan_interval=_read(fi, start_byte + 29, 1, 'binary'),
        nbr_decimation_filter_coef=_read(fi, start_byte + 30, 2, 'binary'),
    )
    return extended_header_2


def _read_extended_header_3(fi, start_byte):
    """
    Extract information contained in the extended header block number 3.
    """
    extended_header_3 = dict(
        receiver_line_number=_read(fi, start_byte, 4, 'binary'),
        receiver_point=_read(fi, start_byte + 4, 4, 'binary'),
        receiver_point_index=_read(fi, start_byte + 8, 1, 'binary'),
        first_shot_line=_read(fi, start_byte + 9, 4, 'binary'),
        first_shot_point=_read(fi, start_byte + 13, 4, 'binary'),
        first_shot_point_index=_read(fi, start_byte + 17, 1, 'binary'),
        last_shot_line=_read(fi, start_byte + 18, 4, 'binary'),
        last_shot_point=_read(fi, start_byte + 22, 4, 'binary'),
        last_shot_point_index=_read(fi, start_byte + 26, 1, 'binary'),
    )
    return extended_header_3


def _read_extended_header(fi, start_byte, block_number, nbr_coeff):
    """
    Extract information contained in the ith extended header block (i>3).
    """
    extended_header = {}
    for i in range(0, nbr_coeff):
        key = 'coeff_' + str(i + 1)
        extended_header[key] = _read(fi, start_byte, 4, 'IEEE')
        start_byte += 4
    return extended_header


if __name__ == '__main__':
    import doctest
    doctest.testmod(exclude_empty=True)



# -*- coding: utf-8 -*-
"""
REFTEK130 read support, core routines.
"""
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)
from future.builtins import *  # NOQA
from future.utils import native_str

import copy
import io
import os
import warnings

import numpy as np

from obspy import Trace, Stream, UTCDateTime
from obspy.core.util.obspy_types import ObsPyException

from .packet import (Packet, EHPacket, _initial_unpack_packets, PACKET_TYPES,
                     PACKET_TYPES_IMPLEMENTED, PACKET_FINAL_DTYPE,
                     Reftek130UnpackPacketError, _unpack_C0_C2_data)


NOW = UTCDateTime()


class Reftek130Exception(ObsPyException):
    pass


def _is_reftek130(filename):
    """
    Checks whether a file is REFTEK130 format or not.

    :type filename: str
    :param filename: REFTEK130 file to be checked.
    :rtype: bool
    :return: ``True`` if a REFTEK130 file.

    Checks if overall length of file is consistent (i.e. multiple of 1024
    bytes) and checks for valid packet type identifiers in the first 20
    expected packet positions.
    """
    if not os.path.isfile(filename):
        return False
    filesize = os.stat(filename).st_size
    # check if overall file size is a multiple of 1024
    if filesize < 1024 or filesize % 1024 != 0:
        return False

    with open(filename, 'rb') as fp:
        # check first 20 expected packets' type header field
        for i in range(20):
            packet_type = fp.read(2).decode("ASCII", "replace")
            if not packet_type:
                # reached end of file..
                break
            if packet_type not in PACKET_TYPES:
                return False
            fp.seek(1022, os.SEEK_CUR)
    return True


def _read_reftek130(filename, network="", location="", component_codes=None,
                    headonly=False, verbose=False,
                    sort_permuted_package_sequence=False, **kwargs):
    """
    Read a REFTEK130 file into an ObsPy Stream.

    :type filename: str
    :param filename: REFTEK130 file to be read.
    :type network: str
    :param network: Network code to fill in for all data (network code is not
        stored in EH/ET/DT packets).
    :type location: str
    :param location: Location code to fill in for all data (network code is not
        stored in EH/ET/DT packets).
    :type component_codes: list
    :param component_codes: Iterable of single-character component codes (e.g.
        ``['Z', 'N', 'E']``) to be appended to two-character stream name parsed
        from event header packet (e.g. ``'HH'``) for each of the channels in
        the data (e.g. to make the channel codes in a three channel data file
        to ``'HHZ'``, ``'HHN'``, ``'HHE'`` in the created stream object).
    :type headonly: bool
    :param headonly: Determines whether or not to unpack the data or just
        read the headers.
    :type sort_permuted_package_sequence: bool
    :param sort_permuted_package_sequence: Determines whether or not the
        package list is sorted when a permuted package sequence is encountered.
        This should only be used if problems occur with files that have a
        permuted package sequence (showing the related warning message).
    :rtype: :class:`~obspy.core.stream.Stream`
    """
    # Reftek 130 data format stores only the last two digits of the year.  We
    # currently assume that 00-49 are years 2000-2049 and 50-99 are years
    # 2050-2099. We deliberately raise an exception if the current year will
    # become 2050 (just in case someone really still uses this code then.. ;-)
    # At that point the year would probably have to be explicitly specified
    # when reading data to be on the safe side.
    if NOW.year > 2050:
        raise NotImplementedError()
    try:
        rt130 = Reftek130.from_file(filename)
        st = rt130.to_stream(
            network=network, location=location,
            component_codes=component_codes, headonly=headonly,
            verbose=verbose,
            sort_permuted_package_sequence=sort_permuted_package_sequence)
        st.merge(-1)
        st.sort()
        return st
    except Reftek130UnpackPacketError:
        msg = ("Unable to read file '{}' as a Reftek130 file. Please contact "
               "developers if you think this is a valid Reftek130 file.")
        raise Reftek130Exception(msg.format(filename))


class Reftek130(object):
    _info_header = "Reftek130 ({:d} packets{})"
    _info_compact_header = [
        "Packet Sequence  Byte Count  Data Fmt  Sampling Rate      Time",
        "  | Packet Type   |  Event #  | Station | Channel #         |",
        "  |   |  Unit ID  |    | Data Stream #  |   |  # of samples |",
        "  |   |   |  Exper.#   |   |  |  |      |   |    |          |"]
    _info_compact_footer = ("(detailed packet information with: "
                            "'print(Reftek130.__str__(compact=False))')")

    def __init__(self):
        self._data = np.array([], dtype=PACKET_FINAL_DTYPE)
        self._filename = None

    def __str__(self, compact=True):
        filename = self._filename and ', file: {}'.format(self._filename) or ''
        info = [self._info_header.format(len(self._data), filename)]
        if compact:
            info += copy.deepcopy(self._info_compact_header)
            info[0] = info[0].format(len(self._data))
            for data in self._data:
                info.append(Packet.from_data(data).__str__(compact=True))
            info.append(self._info_compact_footer)
        else:
            for data in self._data:
                info.append(str(Packet.from_data(data)))
        return "\n".join(info)

    @staticmethod
    def from_file(filename):
        with io.open(filename, "rb") as fh:
            string = fh.read()
        rt = Reftek130()
        rt._data = _initial_unpack_packets(string)
        rt._filename = filename
        return rt

    def check_packet_sequence_and_sort(self, sort_permuted_package_sequence):
        """
        Checks if packet sequence is ordered. If not, shows a warning and sorts
        packets by packet sequence if ``sort_permuted_package_sequence=True``.
        This should ensure that data (DT) packets are properly enclosed by the
        appropriate event header/trailer (EH/ET) packets.
        """
        diff = np.diff(self._data['packet_sequence'].astype(np.int16))
        # rollover from 9999 to 0 is not a packet sequence jump..
        jump = (diff < 1) & (diff != -9999)
        if np.any(jump):
            msg = ("Detected permuted packet sequence, sorting.")
            warnings.warn(msg)
            if sort_permuted_package_sequence:
                self._data.sort(order=[
                    native_str(key) for key in ("packet_sequence", "time")])

    def check_packet_sequence_contiguous(self):
        """
        Checks if packet sequence is contiguous, i.e. without missing packets
        in between. Currently raises if that is the case because this case is
        not covered by test data yet.
        """
        diff = np.diff(self._data['packet_sequence'].astype(np.int16))
        if np.any(diff > 1):
            msg = ("Detected a non-contiguous packet sequence!")
            warnings.warn(msg)

    def drop_not_implemented_packet_types(self):
        """
        Checks if there are packets of a type that is currently not implemented
        and drop them showing a warning message.
        """
        is_implemented = np.in1d(
            self._data['packet_type'],
            [x.encode() for x in PACKET_TYPES_IMPLEMENTED])
        # if all packets are of a type that is implemented, the nothing to do..
        if np.all(is_implemented):
            return
        # otherwise reduce packet list to what is implemented and warn
        not_implemented = np.invert(is_implemented)
        count_not_implemented = not_implemented.sum()
        types_not_implemented = np.unique(
            self._data['packet_type'][not_implemented])
        msg = ("Encountered some packets of types that are not "
               "implemented yet (types: {}). Dropped {:d} packets "
               "overall.")
        msg = msg.format(types_not_implemented.tolist(),
                         count_not_implemented)
        warnings.warn(msg)
        self._data = self._data[is_implemented]

    def to_stream(self, network="", location="", component_codes=None,
                  headonly=False, verbose=False,
                  sort_permuted_package_sequence=False):
        """
        :type headonly: bool
        :param headonly: Determines whether or not to unpack the data or just
            read the headers.
        """
        if verbose:
            print(self)
        if not len(self._data):
            msg = "No packet data in Reftek130 object (file: {})"
            raise Reftek130Exception(msg.format(self._filename))
        self.check_packet_sequence_and_sort(sort_permuted_package_sequence)
        self.check_packet_sequence_contiguous()
        self.drop_not_implemented_packet_types()
        if not len(self._data):
            msg = ("No packet data left in Reftek130 object after dropping "
                   "non-implemented packets (file: {})").format(self._filename)
            raise Reftek130Exception(msg)
        st = Stream()
        for event_number in np.unique(self._data['event_number']):
            data = self._data[self._data['event_number'] == event_number]
            # we should have exactly one EH and one ET packet, truncated data
            # sometimes misses the header or trailer packet.
            eh_packets = data[data['packet_type'] == b"EH"]
            et_packets = data[data['packet_type'] == b"ET"]
            if len(eh_packets) == 0 and len(et_packets) == 0:
                msg = ("Reftek data contains data packets without "
                       "corresponding header or trailer packet.")
                raise Reftek130Exception(msg)
            if len(eh_packets) > 1 or len(et_packets) > 1:
                msg = ("Reftek data contains data packets with multiple "
                       "corresponding header or trailer packets.")
                raise Reftek130Exception(msg)
            if len(eh_packets) != 1:
                msg = ("No event header (EH) packets in packet sequence. "
                       "File might be truncated.")
                warnings.warn(msg)
            if len(et_packets) != 1:
                msg = ("No event trailer (ET) packets in packet sequence. "
                       "File might be truncated.")
                warnings.warn(msg)
            # use either the EH or ET packet, they have the same content (only
            # trigger stop time is not in EH)
            if len(eh_packets):
                eh = EHPacket(eh_packets[0])
            else:
                eh = EHPacket(et_packets[0])
            # only C0, C2, 16, 32 encodings supported right now
            if eh.data_format == b"C0":
                encoding = 'C0'
            elif eh.data_format == b"C2":
                encoding = 'C2'
            elif eh.data_format == b"16":
                msg = ("Reftek130 encoding '16' is implemented but untested, "
                       "please provide example data for testing at "
                       "https://github.com/obspy/obspy/issues/new.")
                warnings.warn(msg)
                encoding = '16'
            elif eh.data_format == b"32":
                encoding = '32'
            else:
                msg = ("Reftek data encoding '{}' not implemented yet. Please "
                       "open an issue on GitHub and provide a small (< 50kb) "
                       "test file.").format(eh.data_format)
                raise NotImplementedError(msg)
            header = {
                "network": network,
                "station": (eh.station_name +
                            eh.station_name_extension).strip(),
                "location": location, "sampling_rate": eh.sampling_rate,
                "reftek130": eh._to_dict()}
            delta = 1.0 / eh.sampling_rate
            delta_nanoseconds = int(delta * 1e9)
            inds_dt = data['packet_type'] == b"DT"
            data_channels = np.unique(data[inds_dt]['channel_number'])
            for channel_number in data_channels:
                inds = data['channel_number'] == channel_number
                # channel number of EH/ET packets also equals zero (one of the
                # three unused bytes in the extended header of EH/ET packets)
                inds &= data['packet_type'] == b"DT"
                packets = data[inds]

                # split into contiguous blocks, i.e. find gaps. packet sequence
                # was sorted already..
                endtimes = (
                    packets[:-1]["time"] +
                    packets[:-1]["number_of_samples"].astype(np.int64) *
                    delta_nanoseconds)
                # check if next starttime matches seamless to last chunk
                # 1e-3 seconds == 1e6 nanoseconds is the smallest time
                # difference reftek130 format can represent, so anything larger
                # or equal means a gap/overlap.
                time_diffs_milliseconds_abs = np.abs(
                    packets[1:]["time"] - endtimes) / 1000000
                gaps = time_diffs_milliseconds_abs >= 1
                if np.any(gaps):
                    gap_split_indices = np.nonzero(gaps)[0] + 1
                    contiguous = np.array_split(packets, gap_split_indices)
                else:
                    contiguous = [packets]

                for packets_ in contiguous:
                    starttime = packets_[0]['time']

                    if headonly:
                        sample_data = np.array([], dtype=np.int32)
                        npts = packets_["number_of_samples"].sum()
                    else:
                        if encoding in ('C0', 'C2'):
                            sample_data = _unpack_C0_C2_data(packets_,
                                                             encoding)
                        elif encoding in ('16', '32'):
                            dtype = {'16': np.int16, '32': np.int32}[encoding]
                            # just fix endianness and use correct dtype
                            sample_data = np.require(
                                packets_['payload'],
                                requirements=['C_CONTIGUOUS'])
                            # either int16 or int32
                            sample_data = sample_data.flatten().view(dtype)
                            # switch endianness, rt130 stores in big endian
                            sample_data = sample_data.byteswap()
                        npts = len(sample_data)

                    tr = Trace(data=sample_data, header=copy.deepcopy(header))
                    # channel number is not included in the EH/ET packet
                    # payload, so add it to stats as well..
                    tr.stats.reftek130['channel_number'] = channel_number
                    if headonly:
                        tr.stats.npts = npts
                    tr.stats.starttime = UTCDateTime(ns=starttime)
                    # if component codes were explicitly provided, use them
                    # together with the stream label
                    if component_codes is not None:
                        tr.stats.channel = (eh.stream_name.strip() +
                                            component_codes[channel_number])
                    # otherwise check if channel code is set for the given
                    # channel (seems to be not the case usually)
                    elif eh.channel_code[channel_number] is not None:
                        tr.stats.channel = eh.channel_code[channel_number]
                    # otherwise fall back to using the stream label together
                    # with the number of the channel in the file (starting with
                    # 0, as Z-1-2 is common use for data streams not oriented
                    # against North)
                    else:
                        msg = ("No channel code specified in the data file "
                               "and no component codes specified. Using "
                               "stream label and number of channel in file as "
                               "channel codes.")
                        warnings.warn(msg)
                        tr.stats.channel = (
                            eh.stream_name.strip() + str(channel_number))
                    # check if endtime of trace is consistent
                    t_last = packets_[-1]['time']
                    npts_last = packets_[-1]['number_of_samples']
                    try:
                        if not headonly:
                            assert npts == len(sample_data)
                        if npts_last:
                            assert tr.stats.endtime == UTCDateTime(
                                ns=t_last) + (npts_last - 1) * delta
                        if npts:
                            assert tr.stats.endtime == (
                                tr.stats.starttime + (npts - 1) * delta)
                    except AssertionError:
                        msg = ("Reftek file has a trace with an inconsistent "
                               "endtime or number of samples. Please open an "
                               "issue on GitHub and provide your file for"
                               "testing.")
                        raise Reftek130Exception(msg)
                    st += tr

        return st


if __name__ == '__main__':
    import doctest
    doctest.testmod(exclude_empty=True)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
seiscomp.core test suite.

:author:
    EOST (École et Observatoire des Sciences de la Terre)
:copyright:
    The ObsPy Development Team (devs@obspy.org)
:license:
    GNU Lesser General Public License, Version 3
    (https://www.gnu.org/copyleft/lesser.html)
"""
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)
from future.builtins import *  # NOQA

import os
import unittest

from obspy.io.seiscomp.core import _is_sc3ml, validate


class CoreTestCase(unittest.TestCase):
    """
    Test suite for obspy.io.seiscomp.event
    """
    def setUp(self):
        self.data_dir = os.path.join(os.path.dirname(__file__), 'data')

    def test_sc3ml_versions(self):
        """
        Test multiple schema versions
        """
        for version in ['0.3', '0.5', '0.9', '0.10']:
            filename = os.path.join(self.data_dir, 'version%s' % version)
            self.assertTrue(_is_sc3ml(filename))

    def test_sc3ml_no_version_attribute(self):
        filename = os.path.join(self.data_dir, 'no_version_attribute.sc3ml')
        self.assertTrue(_is_sc3ml(filename))

    def test_validate(self):
        filename = os.path.join(self.data_dir, 'qml-example-1.2-RC3.sc3ml')
        self.assertTrue(validate(filename))
        self.assertFalse(validate(filename, version='0.8'))

        with self.assertRaises(ValueError) as e:
            validate(filename, version='0.11')

        expected_error = ("0.11 is not a supported version. Use one of these "
                          "versions: [0.3, 0.5, 0.6, 0.7, 0.8, 0.9, 0.10].")
        self.assertEqual(e.exception.args[0], expected_error)


def suite():
    return unittest.makeSuite(CoreTestCase, "test")


if __name__ == '__main__':
    unittest.main(defaultTest='suite')



# -*- coding: utf-8 -*-
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)
from future.builtins import *  # NOQA

from .blockette import Blockette
from ..fields import Integer, VariableString


class Blockette051(Blockette):
    """
    Blockette 051: Station Comment Blockette.

    Sample:
    05100351992,001~1992,002~0740000000
    """

    id = 51
    name = "Station Comment"
    fields = [
        VariableString(3, "Beginning effective time", 1, 22, 'T'),
        VariableString(4, "End effective time", 0, 22, 'T', optional=True),
        Integer(5, "Comment code key", 4, xpath=31),
        Integer(6, "Comment level", 6, ignore=True)
    ]



"""
obspy.io.dmc - INGV DMX file format reader for ObsPy
====================================================

Functions to read waveform data from the standard INGV DMX format.

:author:
    Thomas Lecocq
    Andrea Cannatta
:copyright:
    The ObsPy Development Team (devs@obspy.org)
:license:
    GNU Lesser General Public License, Version 3
    (https://www.gnu.org/copyleft/lesser.html)


Reading the waveforms
---------------------
Reading DMX is handled by using ObsPy's standard
:func:`~obspy.core.stream.read` function. The format can be detected
automatically, however setting the ``format`` parameter as "DMX" lead to a
speed up.
One optional keyword argument is available: ``station``. It is automatically
passed to the :func:`obspy.io.dmx.core._read_dmx`. Its format 

>>> from obspy import read
>>> # these two are equivalent, but the second case should be faster:
>>> st = read("/path/to/131114_090600.dmx")
>>> print(st)  #doctest: +NORMALIZE_WHITESPACE
2 Trace(s) in Stream:
ETNA.EMFO..Z | 2013-11-14T09:06:00.000000Z - 2013-11-14T09:06:59.990000Z | 100.0 Hz, 6000 samples
ETNA.EMPL..Z | 2013-11-14T09:06:00.000000Z - 2013-11-14T09:06:59.990000Z | 100.0 Hz, 6000 samples

>>> st = read("/path/to/131114_090600.dmx", format='DMX')
>>> print(st)  #doctest: +NORMALIZE_WHITESPACE
2 Trace(s) in Stream:
ETNA.EMFO..Z | 2013-11-14T09:06:00.000000Z - 2013-11-14T09:06:59.990000Z | 100.0 Hz, 6000 samples
ETNA.EMPL..Z | 2013-11-14T09:06:00.000000Z - 2013-11-14T09:06:59.990000Z | 100.0 Hz, 6000 samples

If the file is very large and only one station code needs to be fetched,
using the ``station`` parameter may speed the reading process:

>>> st = read("/path/to/131114_090600.dmx", station="EMFO")
>>> print(st)  # doctest: +ELLIPSIS
1 Trace(s) in Stream:
ETNA.EMFO..Z | 2013-11-14T09:06:00.000000Z - 2013-11-14T09:06:59.990000Z | 100.0 Hz, 6000 samples

"""
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)
from future.builtins import *  # NOQA


if __name__ == '__main__':
    import doctest
    doctest.testmod(exclude_empty=True)



# -*- coding: utf-8 -*-
"""
The obspy.imaging.spectrogram test suite.
"""
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)
from future.builtins import *  # NOQA

import os
import unittest
import warnings

import numpy as np

from obspy import Stream, Trace, UTCDateTime
from obspy.core.util.testing import ImageComparison
from obspy.imaging import spectrogram


class SpectrogramTestCase(unittest.TestCase):
    """
    Test cases for spectrogram plotting.
    """
    def setUp(self):
        # directory where the test files are located
        self.path = os.path.join(os.path.dirname(__file__), 'images')

    def test_spectrogram(self):
        """
        Create spectrogram plotting examples in tests/output directory.
        """
        # Create dynamic test_files to avoid dependencies of other modules.
        # set specific seed value such that random numbers are reproducible
        np.random.seed(815)
        head = {
            'network': 'BW', 'station': 'BGLD',
            'starttime': UTCDateTime(2007, 12, 31, 23, 59, 59, 915000),
            'sampling_rate': 200.0, 'channel': 'EHE'}
        tr = Trace(data=np.random.randint(0, 1000, 824), header=head)
        st = Stream([tr])
        # 1 - using log=True
        with ImageComparison(self.path, 'spectrogram_log.png') as ic:
            with warnings.catch_warnings(record=True):
                warnings.resetwarnings()
                with np.errstate(all='warn'):
                    spectrogram.spectrogram(
                        st[0].data, log=True, outfile=ic.name,
                        samp_rate=st[0].stats.sampling_rate, show=False)
        # 2 - using log=False
        with ImageComparison(self.path, 'spectrogram.png') as ic:
            spectrogram.spectrogram(st[0].data, log=False, outfile=ic.name,
                                    samp_rate=st[0].stats.sampling_rate,
                                    show=False)


def suite():
    return unittest.makeSuite(SpectrogramTestCase, 'test')


if __name__ == '__main__':
    unittest.main(defaultTest='suite')



# -*- coding: utf-8 -*-
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)
from future.builtins import *  # NOQA
from future.utils import native_str

import io
import json
import os
import unittest
import warnings

from obspy.io.json.default import Default
from obspy.io.json.core import get_dump_kwargs, _write_json
from obspy.io.quakeml.core import _read_quakeml


warnings.filterwarnings("ignore")


class JSONTestCase(unittest.TestCase):
    """Test JSON module classes and functions"""
    def setUp(self):
        self.path = os.path.join(os.path.dirname(__file__))
        qml_file = os.path.join(self.path, "..", "..", "quakeml", "tests",
                                "data", "qml-example-1.2-RC3.xml")
        self.c = _read_quakeml(qml_file)
        self.event = self.c.events[0]

    def verify_json(self, s):
        """Test an output is a string and is JSON"""
        self.assertTrue(isinstance(s, (str, native_str)))
        j = json.loads(s)
        self.assertTrue(isinstance(j, dict))

    def test_default(self):
        """Test Default function class"""
        default = Default()
        self.assertTrue(hasattr(default, '_catalog_attrib'))
        self.assertTrue(hasattr(default, 'OMIT_NULLS'))
        self.assertTrue(hasattr(default, 'TIME_FORMAT'))
        self.assertTrue(hasattr(default, '__init__'))
        self.assertTrue(hasattr(default, '__call__'))
        s = json.dumps(self.event, default=default)
        self.verify_json(s)

    def test_get_dump_kwargs(self):
        """Test getting kwargs for json.dumps"""
        kw = get_dump_kwargs()
        self.assertIn('default', kw)
        self.assertIn('separators', kw)
        self.assertTrue(isinstance(kw['default'], Default))
        self.assertTrue(kw['default'].OMIT_NULLS)
        self.assertEqual(kw['separators'], (',', ':'))
        s1 = json.dumps(self.event, **kw)
        self.verify_json(s1)
        kw = get_dump_kwargs(minify=False, no_nulls=False)
        self.assertIn('default', kw)
        self.assertNotIn('separators', kw)
        self.assertTrue(isinstance(kw['default'], Default))
        self.assertFalse(kw['default'].OMIT_NULLS)
        s2 = json.dumps(self.event, **kw)
        self.verify_json(s2)
        # Compacted version is smaller
        self.assertTrue(len(s1) < len(s2))

    def test_write_json(self):
        memfile = io.StringIO()
        _write_json(self.c, memfile)
        memfile.seek(0, 0)
        # Verify json module can load
        j = json.load(memfile)
        self.assertTrue(isinstance(j, dict))
        # Test registered method call
        memfile = io.StringIO()
        self.c.write(memfile, format="json")
        memfile.seek(0, 0)
        # Verify json module can load
        j = json.load(memfile)
        self.assertTrue(isinstance(j, dict))

    def tearDown(self):
        del self.event


def suite():
    return unittest.makeSuite(JSONTestCase, 'test')


if __name__ == '__main__':
    unittest.main(defaultTest='suite')



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------
#  Filename: unpack.py
#  Purpose: Routines for unpacking SEG Y data formats.
#   Author: Lion Krischer
#    Email: krischer@geophysik.uni-muenchen.de
#
# Copyright (C) 2010 Lion Krischer
# --------------------------------------------------------------------
"""
Functions that will all take a file pointer and the sample count and return a
NumPy array with the unpacked values.
"""
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)
from future.builtins import *  # NOQA
from future.utils import native_str

import ctypes as C  # NOQA
import os
import sys
import warnings

import numpy as np

from obspy.core.compatibility import from_buffer
from .util import clibsegy


# Get the system byte order.
BYTEORDER = sys.byteorder
if BYTEORDER == 'little':
    BYTEORDER = '<'
else:
    BYTEORDER = '>'


clibsegy.ibm2ieee.argtypes = [
    np.ctypeslib.ndpointer(dtype=np.float32, ndim=1,
                           flags=native_str('C_CONTIGUOUS')),
    C.c_int]
clibsegy.ibm2ieee.restype = C.c_void_p


def unpack_4byte_ibm(file, count, endian='>'):
    """
    Unpacks 4 byte IBM floating points.
    """
    # Read as 4 byte integer so bit shifting works.
    data = from_buffer(file.read(count * 4), dtype=np.float32)
    # Swap the byte order if necessary.
    if BYTEORDER != endian:
        data = data.byteswap()
    length = len(data)
    # Call the C code which transforms the data inplace.
    clibsegy.ibm2ieee(data, length)
    return data


# Old pure Python/NumPy code
#
# def unpack_4byte_ibm(file, count, endian='>'):
#    """
#    Unpacks 4 byte IBM floating points.
#    """
#    # Read as 4 byte integer so bit shifting works.
#    data = np.fromstring(file.read(count * 4), dtype=np.int32)
#    # Swap the byte order if necessary.
#    if BYTEORDER != endian:
#        data = data.byteswap()
#    # See https://mail.scipy.org/pipermail/scipy-user/2009-January/019392.html
#    # XXX: Might need check for values out of range:
#    # https://bytes.com/topic/c/answers/
#    #         221981-c-code-converting-ibm-370-floating-point-ieee-754-a
#    sign = np.bitwise_and(np.right_shift(data, 31), 0x01)
#    exponent = np.bitwise_and(np.right_shift(data, 24), 0x7f)
#    mantissa = np.bitwise_and(data, 0x00ffffff)
#    # Force single precision.
#    mantissa = np.require(mantissa, 'float32')
#    mantissa /= 0x1000000
#    # Do the following calculation in a weird way to avoid autocasting to
#    # float64.
#    # data = (1.0 - 2.0 * sign) * mantissa * 16.0 ** (exponent - 64.0)
#    sign *= -2.0
#    sign += 1.0
#    mantissa *= 16.0 ** (exponent - 64)
#    mantissa *= sign
#    return mantissa


def unpack_4byte_integer(file, count, endian='>'):
    """
    Unpacks 4 byte integers.
    """
    # Read as 4 byte integer so bit shifting works.
    data = from_buffer(file.read(count * 4), dtype=np.int32)
    # Swap the byte order if necessary.
    if BYTEORDER != endian:
        data = data.byteswap()
    return data


def unpack_2byte_integer(file, count, endian='>'):
    """
    Unpacks 2 byte integers.
    """
    # Read as 4 byte integer so bit shifting works.
    data = from_buffer(file.read(count * 2), dtype=np.int16)
    # Swap the byte order if necessary.
    if BYTEORDER != endian:
        data = data.byteswap()
    return data


def unpack_4byte_fixed_point(file, count, endian='>'):
    raise NotImplementedError


def unpack_4byte_ieee(file, count, endian='>'):
    """
    Unpacks 4 byte IEEE floating points.
    """
    # Read as 4 byte integer so bit shifting works.
    data = from_buffer(file.read(count * 4), dtype=np.float32)
    # Swap the byte order if necessary.
    if BYTEORDER != endian:
        data = data.byteswap()
    return data


def unpack_1byte_integer(file, count, endian='>'):
    raise NotImplementedError


class OnTheFlyDataUnpacker:
    """
    Tie-up a data sample unpack function with its parameters.

    This class allows for data to be read directly from the disk as needed,
    preventing the need to store data in memory.
    """
    def __init__(self, unpack_function, filename, filemode, seek, count,
                 endian='>'):
        self.unpack_function = unpack_function
        self.filename = filename
        self.filemode = filemode
        self.seek = seek
        self.count = count
        self.endian = endian
        self.mtime = os.path.getmtime(self.filename)

    def __call__(self):
        mtime = os.path.getmtime(self.filename)
        if mtime != self.mtime:
            msg = "File '%s' changed since reading headers" % self.filename
            msg += "; data may be read incorrectly "
            msg += "(modification time = %s)." % mtime
            warnings.warn(msg)
        with open(self.filename, self.filemode) as fp:
            fp.seek(self.seek)
            raw = self.unpack_function(fp, self.count, endian=self.endian)
        return raw



# -*- coding: utf-8 -*-
"""
obspy.clients.arclink - ArcLink/WebDC request client for ObsPy
==============================================================

ArcLink is a distributed data request protocol usable to access archived
waveform data in the MiniSEED or SEED format and associated meta information as
Dataless SEED files. It has been originally founded within the German WebDC
initiative of GEOFON_ (Geoforschungsnetz) and BGR_ (Bundesanstalt für
Geowissenschaften und Rohstoffe). ArcLink has been designed as a "straight
consequent continuation" of the NetDC concept originally developed by the IRIS_
DMC. Instead of requiring waveform data via E-mail or FTP requests, ArcLink
offers a direct TCP/IP communication approach. A prototypic web-based request
tool is available via the WebDC homepage at http://www.webdc.eu.

Recent development efforts within the NERIES_ (Network of Excellence of
Research and Infrastructures for European Seismology) project focuses on
extending the ArcLink network to all major seismic data centers within Europe
in order to create an European Integrated Data Center (EIDAC). Currently
(September 2009) there are four European data centers contributing to this
network: ORFEUS_, GFZ_ (GeoForschungsZentrum), INGV_ (Istituto Nazionale di
Geofisica e Vulcanologia), and IPGP_ (Institut de Physique du Globe de Paris).

:copyright:
    The ObsPy Development Team (devs@obspy.org)
:license:
    GNU Lesser General Public License, Version 3
    (https://www.gnu.org/copyleft/lesser.html)

Basic Usage
-----------

.. note::
    The default client needs to open port 18002 to the host webdc.eu via TCP/IP
    in order to download the requested data. Please make sure that no firewall
    is blocking access to this server/port combination.

.. note::
    The ``user`` keyword in the following examples is used for identification
    with the ArcLink server as well as for usage statistics within the data
    center, so please provide a meaningful user id such as an email address.

(1) :meth:`~obspy.clients.arclink.client.Client.get_waveforms()`: The following
    example illustrates how to request and plot 18 seconds of all three single
    band channels (``"EH*"``) of station Jochberg/Hochstaufen (``"RJOB"``)
    of the Bavarian network (``"BW"``) for an seismic event around
    2009-08-20 04:03:12 (UTC).

    >>> from obspy import UTCDateTime
    >>> from obspy.clients.arclink.client import Client
    >>> client = Client(user='test@obspy.org')
    >>> t = UTCDateTime("2009-08-20 04:03:12")
    >>> st = client.get_waveforms("BW", "RJOB", "", "EH*", t - 3, t + 15)
    >>> st.plot()  # doctest: +SKIP

    Waveform data fetched from an ArcLink node is converted into an ObsPy
    :class:`~obspy.core.stream.Stream` object. The seismogram is truncated by
    the ObsPy client to the actual requested time span, as ArcLink internally
    cuts SEED files for performance reasons on record base in order to avoid
    uncompressing the waveform data. The output of the script above is shown in
    the next picture.

    .. plot::

        from obspy import UTCDateTime
        from obspy.clients.arclink.client import Client
        client = Client(user='test@obspy.org')
        t = UTCDateTime("2009-08-20 04:03:12")
        st = client.get_waveforms("BW", "RJOB", "", "EH*", t - 3, t + 15)
        st.plot()  # doctest: +SKIP

(2) :meth:`~obspy.clients.arclink.client.Client.get_paz()`: Requests poles,
    zeros, gain and sensitivity of a single channel at a given time.

    >>> from obspy import UTCDateTime
    >>> from obspy.clients.arclink.client import Client
    >>> client = Client(user='test@obspy.org')
    >>> dt = UTCDateTime(2009, 1, 1)
    >>> paz = client.get_paz('BW', 'MANZ', '', 'EHZ', dt)
    >>> paz  # doctest: +NORMALIZE_WHITESPACE +SKIP
    AttribDict({'poles': [(-0.037004+0.037016j), (-0.037004-0.037016j),
                          (-251.33+0j), (-131.04-467.29j), (-131.04+467.29j)],
                'sensitivity': 2516778600.0,
                'zeros': [0j, 0j],
                'name': 'LMU:STS-2/N/g=1500',
                'gain': 60077000.0})

(3) :meth:`~obspy.clients.arclink.client.Client.save_response()`: Writes
    response information into a file.

    >>> from obspy import UTCDateTime
    >>> from obspy.clients.arclink.client import Client
    >>> client = Client(user='test@obspy.org')
    >>> t = UTCDateTime(2009, 1, 1)
    >>> client.save_response('BW.MANZ..EHZ.dataless', 'BW', 'MANZ', '', '*',
    ...                     t, t + 1, format="SEED")  # doctest: +SKIP

(4) :meth:`~obspy.clients.arclink.client.Client.save_waveforms()`: Writes the
    requested waveform unmodified into your local file system. Here we request
    a Full SEED volume.

    >>> from obspy import UTCDateTime
    >>> from obspy.clients.arclink.client import Client
    >>> client = Client(user='test@obspy.org')
    >>> t = UTCDateTime(2009, 1, 1, 12, 0)
    >>> client.save_waveforms('BW.MANZ..EHZ.seed', 'BW', 'MANZ', '', '*',
    ...                     t, t + 20, format='FSEED')  # doctest: +SKIP

(5) :meth:`~obspy.clients.arclink.client.Client.get_inventory()`: Request
    inventory data.

    >>> from obspy import UTCDateTime
    >>> from obspy.clients.arclink.client import Client
    >>> client = Client(user='test@obspy.org')
    >>> inv = client.get_inventory('BW', 'M*', '*', 'EHZ', restricted=False,
    ...                           permanent=True, min_longitude=12,
    ...                           max_longitude=12.2) #doctest: +SKIP
    >>> inv.keys()  # doctest: +SKIP
    ['BW.MROB', 'BW.MANZ..EHZ', 'BW', 'BW.MANZ', 'BW.MROB..EHZ']
    >>> inv['BW']  # doctest: +SKIP
    AttribDict({'description': 'BayernNetz', 'region': 'Germany', ...
    >>> inv['BW.MROB']  # doctest: +SKIP
    AttribDict({'code': 'MROB', 'description': 'Rosenbuehl, Bavaria', ...

Further Resources
-----------------
* ArcLink protocol specifications:

  * https://www.seiscomp3.org/wiki/doc/applications/arclink
  * http://geofon.gfz-potsdam.de/_sc3_neries_/arclink.pdf
  * https://raw.github.com/obspy/obspy/master/obspy/clients/arclink/docs/\
protocol.txt

* `Short introduction <http://www.webdc.eu/arclink/help.html>`_ to the ArcLink
  protocol
* Latest `ArcLink server
  <ftp://ftp.gfz-potsdam.de/pub/home/st/GEOFON/software/SeisComP/ArcLink/>`_
* `SeismoLink <http://neriesdataportalblog.freeflux.net/webservices/>`_: a SOAP
  Web service on top of the ArcLink network

.. _GEOFON:
        http://geofon.gfz-potsdam.de
.. _BGR:
        https://www.bgr.de
.. _IRIS:
        https://www.iris.edu/hq/
.. _NERIES:
        http://www.neries-eu.org
.. _INGV:
        http://www.ingv.it
.. _ORFEUS:
        http://www.orfeus-eu.org
.. _GFZ:
        http://www.gfz-potsdam.de
.. _IPGP:
        http://www.ipgp.fr
.. _`GNU Lesser General Public License, Version 3`:
        https://www.gnu.org/copyleft/lesser.html
"""
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)
from future.builtins import *  # NOQA

import os
import warnings

from obspy.core.util.deprecation_helpers import ObsPyDeprecationWarning

# Raise a deprecation warning. This has been requested by the EIDA management
# board.
msg = (
    "The ArcLink protocol will be deprecated in the near future. Please, use "
    "the client contacting the routing service provided by EIDA: "
    "https://docs.obspy.org/packages/obspy.clients.fdsn.html#"
    "basic-routing-clients-usage")
# suppress warning on docs build
if os.environ.get('SPHINX') != 'true':
    warnings.warn(msg, category=ObsPyDeprecationWarning)

from .client import Client  # NOQA


if __name__ == '__main__':
    import doctest
    doctest.testmod(exclude_empty=True)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Non-geographical restrictions and constraints for the mass downloader.

:copyright:
    Lion Krischer (krischer@geophysik.uni-muenchen.de), 2014-2015
:license:
    GNU Lesser General Public License, Version 3
    (https://www.gnu.org/copyleft/lesser.html)
"""
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)
from future.builtins import *  # NOQA
from future.utils import native_str

from obspy.core.compatibility import collections_abc

import obspy


class Restrictions(object):
    """
    Class storing non-domain restrictions for a query. This is best explained
    with two examples. See the list below for a more detailed explanation
    of the parameters. The first set of restrictions is useful for event
    based earthquake set queries.

    >>> import obspy
    >>> restrictions = Restrictions(
    ...     # Get data from 5 minutes before the event to one hour after the
    ...     # event.
    ...     starttime=obspy.UTCDateTime(2012, 1, 1),
    ...     endtime=obspy.UTCDateTime(2012, 1, 2),
    ...     # You might not want to deal with gaps in the data.
    ...     reject_channels_with_gaps=True,
    ...     # And you might only want waveforms that have data for at least
    ...     # 95 % of the requested time span.
    ...     minimum_length=0.95,
    ...     # No two stations should be closer than 10 km to each other.
    ...     minimum_interstation_distance_in_m=10E3,
    ...     # Only HH or BH channels. If a station has HH channels,
    ...     # those will be downloaded, otherwise the BH. Nothing will be
    ...     # downloaded if it has neither.
    ...     channel_priorities=["HH[ZNE]", "BH[ZNE]"],
    ...     # Location codes are arbitrary and there is no rule as to which
    ...     # location is best.
    ...     location_priorities=["", "00", "10"])


    And the restrictions for downloading a noise data set might look similar to
    the following:

    >>> import obspy
    >>> restrictions = Restrictions(
    ...     # Get data for a whole year.
    ...     starttime=obspy.UTCDateTime(2012, 1, 1),
    ...     endtime=obspy.UTCDateTime(2013, 1, 1),
    ...     # Chunk it to have one file per day.
    ...     chunklength_in_sec=86400,
    ...     # Considering the enormous amount of data associated with
    ...     # continuous requests, you might want to limit the data based on
    ...     # SEED identifiers. If the location code is specified, the
    ...     # location priority list is not used; the same is true for the
    ...     # channel argument and priority list.
    ...     network="BW", station="A*", location="", channel="BH*",
    ...     # The typical use case for such a data set are noise correlations
    ...     # where gaps are dealt with at a later stage.
    ...     reject_channels_with_gaps=False,
    ...     # Same is true with the minimum length. Any data during a day
    ...     # might be useful.
    ...     minimum_length=0.0,
    ...     # Sanitize makes sure that each MiniSEED file also has an
    ...     # associated StationXML file, otherwise the MiniSEED files will
    ...     # be deleted afterwards. This is not desirable for large noise
    ...     # data sets.
    ...     sanitize=False,
    ...     # Guard against the same station having different names.
    ...     minimum_interstation_distance_in_m=100.0)

    The ``network``, ``station``, ``location``, and ``channel`` codes are
    directly passed to the `station` service of each fdsn-ws implementation
    and can thus take comma separated string lists as arguments, i.e.

    .. code-block:: python

        restrictions = Restrictions(
            ...
            network="BW,G?", station="A*,B*",
            ...
            )

    Not all fdsn-ws implementations support the direct exclusion of network
    or station codes. The ``exclude_networks`` and ``exclude_stations``
    arguments should thus be used for that purpose to ensure compatibility
    across all data providers, e.g.

    .. code-block:: python

        restrictions = Restrictions(
            ...
            network="B*,G*", station="A*, B*",
            exclude_networks=["BW", "GR"],
            exclude_stations=["AL??", "*O"],
            ...
            )

    It is also possible to restrict the downloaded stations to stations part of
    an existing inventory object which can originate from a StationXML file or
    from other sources. It will only keep stations that are part of the
    inventory object. Channels are still selected dynamically based on the
    other restrictions. Keep in mind that all other restrictions still apply -
    passing an inventory will just further restrict the possibly downloaded
    data.

    .. code-block:: python

        restrictions = Restrictions(
            ...
            limit_stations_to_inventory=inv,
            ...
            )

    :param starttime: The start time of the data to be downloaded.
    :type starttime: :class:`~obspy.core.utcdatetime.UTCDateTime`
    :param endtime: The end time of the data.
    :type endtime: :class:`~obspy.core.utcdatetime.UTCDateTime`
    :param station_starttime: The start time of the station files. If not
        given, the ``starttime`` argument will be used. This is useful when
        trying to incorporate multiple waveform datasets with a central
        station file archive as StationXML files can be downloaded once and
        for the whole time span.
    :type station_starttime: :class:`~obspy.core.utcdatetime.UTCDateTime`
    :param station_endtime: The end time of the station files. Analogous to
        the ``station_starttime`` argument.
    :type station_endtime: :class:`~obspy.core.utcdatetime.UTCDateTime`
    :param chunklength_in_sec: The length of one chunk in seconds. If set,
        the time between ``starttime`` and ``endtime`` will be divided into
        segments of ``chunklength_in_sec`` seconds. Useful for continuous data
        requests. Set to ``None`` if one piece of data is desired between
        ``starttime`` and ``endtime`` (the default).
    :type chunklength_in_sec: float
    :param network: The network code. Can contain wildcards.
    :type network: str
    :param station: The station code. Can contain wildcards.
    :type station: str
    :param location: The location code. Can contain wildcards.
    :type location: str
    :param channel: The channel code. Can contain wildcards.
    :type channel: str
    :param exclude_networks: A list of potentially wildcarded networks that
        should not be downloaded.
    :type exclude_networks: list of str
    :param exclude_stations: A list of potentially wildcarded stations that
        should not be downloaded.
    :type exclude_stations: list of str
    :param limit_stations_to_inventory: If given, only stations part of the
        this inventory object will be downloaded. All other restrictions
        still apply - this just serves to further limit the set of stations
        to download.
    :type limit_stations_to_inventory:
        :class:`~obspy.core.inventory.inventory.Inventory`
    :param reject_channels_with_gaps: If True (default), MiniSEED files with
        gaps and/or overlaps will be rejected.
    :type reject_channels_with_gaps: bool
    :param minimum_length: The minimum length of the data as a fraction of
        the requested time frame. After a channel has been downloaded it
        will be checked that its total length is at least that fraction of
        the requested time span. Will be rejected otherwise. Must be between
        ``0.0`` and ``1.0``, defaults to ``0.9``.
    :type minimum_length: float
    :param sanitize: Sanitize makes sure that each MiniSEED file also has an
         associated StationXML file, otherwise the MiniSEED files will be
         deleted afterwards. This is potentially not desirable for large noise
         data sets.
    :type sanitize: bool
    :param minimum_interstation_distance_in_m: The minimum inter-station
        distance. Data from any new station closer to any existing station
        will not be downloaded. Also used for duplicate station detection as
        sometimes stations have different names for different webservice
        providers. Defaults to `1000 m`.
    :type minimum_interstation_distance_in_m: float
    :param channel_priorities: Priority list for the channels. Will not be
        used if the ``channel`` argument is used.
    :type channel_priorities: list of str
    :param location_priorities: Priority list for the locations. Will not be
        used if the ``location`` argument is used.
    :type location_priorities: list of str
    """
    def __init__(self, starttime, endtime,
                 station_starttime=None, station_endtime=None,
                 chunklength_in_sec=None,
                 network=None, station=None, location=None, channel=None,
                 exclude_networks=tuple(), exclude_stations=tuple(),
                 limit_stations_to_inventory=None,
                 reject_channels_with_gaps=True, minimum_length=0.9,
                 sanitize=True, minimum_interstation_distance_in_m=1000,
                 channel_priorities=("HH[ZNE12]", "BH[ZNE12]",
                                     "MH[ZNE12]", "EH[ZNE12]",
                                     "LH[ZNE12]", "HL[ZNE12]",
                                     "BL[ZNE12]", "ML[ZNE12]",
                                     "EL[ZNE12]", "LL[ZNE12]",
                                     "SH[ZNE12]"),
                 location_priorities=("", "00", "10", "01", "20", "02", "30",
                                      "03", "40", "04", "50", "05", "60",
                                      "06", "70", "07", "80", "08", "90",
                                      "09")):
        # Awkward logic to keep track whether or not the location priorities
        # are equal to the default values. This "solution" keeps the function
        # signature intact and it also located close to where the location
        # priorities are set.
        if location_priorities == (
                "", "00", "10", "01", "20", "02", "30", "03", "40", "04", "50",
                "05", "60", "06", "70", "07", "80", "08", "90", "09"):
            self._loc_prios_are_default_values = True
        else:
            self._loc_prios_are_default_values = False

        self.starttime = obspy.UTCDateTime(starttime)
        self.endtime = obspy.UTCDateTime(endtime)
        self.station_starttime = station_starttime and \
            obspy.UTCDateTime(station_starttime)
        self.station_endtime = station_endtime and \
            obspy.UTCDateTime(station_endtime)
        if self.station_starttime and self.station_starttime > self.starttime:
            raise ValueError("The station start time must be smaller than the "
                             "main start time.")
        if self.station_endtime and self.station_endtime < self.endtime:
            raise ValueError("The station end time must be larger than the "
                             "main end time.")
        self.chunklength = chunklength_in_sec and float(chunklength_in_sec)
        self.network = network
        self.station = station
        self.location = location
        self.channel = channel
        self.exclude_networks = exclude_networks
        self.exclude_stations = exclude_stations
        self.reject_channels_with_gaps = reject_channels_with_gaps
        self.minimum_length = minimum_length
        self.sanitize = bool(sanitize)

        # These must be iterables, but not strings.
        if not isinstance(channel_priorities, collections_abc.Iterable) \
                or isinstance(channel_priorities, (str, native_str)):
            msg = "'channel_priorities' must be a list or other iterable " \
                  "container."
            raise TypeError(msg)

        if not isinstance(location_priorities, collections_abc.Iterable) \
                or isinstance(location_priorities, (str, native_str)):
            msg = "'location_priorities' must be a list or other iterable " \
                  "container."
            raise TypeError(msg)

        self.channel_priorities = channel_priorities
        self.location_priorities = location_priorities

        self.minimum_interstation_distance_in_m = \
            float(minimum_interstation_distance_in_m)

        # Further restrict the possibly downloaded networks and station to
        # the one in the given inventory.
        if limit_stations_to_inventory is not None:
            self.limit_stations_to_inventory = set()
            for net in limit_stations_to_inventory:
                for sta in net:
                    self.limit_stations_to_inventory.add((net.code, sta.code))
        else:
            self.limit_stations_to_inventory = None

    def __eq__(self, other):
        return self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not self == other

    def __iter__(self):
        """
        Iterator yielding time intervals based on the chunklength and
        temporal settings.
        """
        if not self.chunklength:
            return iter([(self.starttime, self.endtime)])

        def it():
            """
            Tiny iterator.
            """
            starttime = self.starttime
            endtime = self.endtime
            chunklength = self.chunklength

            while starttime < endtime:
                yield (starttime, min(starttime + chunklength, endtime))
                starttime += chunklength
            return

        return it()


if __name__ == '__main__':
    import doctest
    doctest.testmod(exclude_empty=True)



# -*- coding: utf-8 -*-
"""
obspy.io.kml - Keyhole Markup Language (KML) write support for ObsPy
====================================================================
This module provides write support for the Keyhole Markup Language (KML)
format.

Write support works via the ObsPy plugin structure for
:class:`~obspy.core.event.Catalog` and
:class:`~obspy.core.inventory.inventory.Inventory`:

>>> from obspy import read_inventory, read_events
>>> inv = read_inventory()  # load example data
>>> inv.write("my_stations.kml", format="KML")  # doctest: +SKIP
>>> cat = read_events()  # load example data
>>> cat.write("my_events.kml", format="KML")  # doctest: +SKIP

For details on further parameters see
:meth:`~obspy.io.kml.core.inventory_to_kml_string` and
:meth:`~obspy.io.kml.core.catalog_to_kml_string`.

:copyright:
    The ObsPy Development Team (devs@obspy.org)
:license:
    GNU Lesser General Public License, Version 3
    (https://www.gnu.org/copyleft/lesser.html)
"""
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)
from future.builtins import *  # NOQA



# -*- coding: utf-8 -*-
"""
The obspy.clients.neic.client test suite.
"""
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)
from future.builtins import *  # NOQA

import unittest

from obspy.core.utcdatetime import UTCDateTime
from obspy.clients.neic import Client


class ClientTestCase(unittest.TestCase):
    """
    Test cases for obspy.clients.neic.client.Client.
    """
    @classmethod
    def setUpClass(cls):
        cls.client = Client(host="137.227.224.97", port=2061, timeout=8)
        cls.starttime = UTCDateTime() - 5 * 60 * 60

    def test_get_waveform(self):
        """
        Tests get_waveforms method. Tests against get_waveforms_nscl method.
        """
        client = self.client
        t = self.starttime
        duration = 1.0
        st = client.get_waveforms_nscl("IUANMO BH.00", t, duration)
        # try a series of requests, compare against get_waveforms_nscl
        args = [["IU", "ANMO", "00", "BH."],
                ["??", "ANMO", "0?", "BH[Z21]"],
                ["IU", "ANM.*", "00", "B??"],
                ["IU", "ANMO", "0*", "BH."],
                ]
        for args_ in args:
            st2 = client.get_waveforms(*args_, starttime=t,
                                       endtime=t + duration)
            self.assertEqual(st, st2)

    def test_get_waveform_nscl(self):
        """
        Tests get_waveforms_nscl method.
        """
        client = self.client
        t = self.starttime
        duration_long = 3600.0
        duration = 1.0
        components = ["1", "2", "Z"]
        # try one longer request to see if fetching multiple blocks works
        st = client.get_waveforms_nscl("IUANMO BH.00", t, duration_long)
        # merge to avoid failing tests simply due to gaps
        st.merge()
        st.sort()
        self.assertEqual(len(st), 3)
        for tr, component in zip(st, components):
            stats = tr.stats
            self.assertEqual(stats.station, "ANMO")
            self.assertEqual(stats.network, "IU")
            self.assertEqual(stats.location, "00")
            self.assertEqual(stats.channel, "BH" + component)
            # requested data duration has some minor fluctuations sometimes but
            # should be pretty close to the expected duration.
            # it should not be over a delta longer than expected (it should be
            # trimmed correctly if more data is returned) but sometimes it's
            # one delta shorter
            self.assertTrue(
                abs(duration_long - (stats.endtime - stats.starttime)) <=
                tr.stats.delta)
            # if the following fails this is likely due to a change at the
            # requested station and simply has to be adapted
            self.assertEqual(stats.sampling_rate, 40.0)
            self.assertEqual(len(tr), 144001)
        # now use shorter piece, this is faster and less error prone (gaps etc)
        st = client.get_waveforms_nscl("IUANMO BH.00", t, duration)
        st.sort()
        # test returned stream
        self.assertEqual(len(st), 3)
        for tr, component in zip(st, components):
            stats = tr.stats
            self.assertEqual(stats.station, "ANMO")
            self.assertEqual(stats.network, "IU")
            self.assertEqual(stats.location, "00")
            self.assertEqual(stats.channel, "BH" + component)
            # requested data duration has some minor fluctuations sometimes but
            # should be pretty close to the expected duration.
            # it should not be over a delta longer than expected (it should be
            # trimmed correctly if more data is returned) but sometimes it's
            # one delta shorter
            self.assertTrue(
                abs(duration - (stats.endtime - stats.starttime)) <=
                tr.stats.delta)
            # if the following fails this is likely due to a change at the
            # requested station and simply has to be adapted
            self.assertEqual(stats.sampling_rate, 40.0)
            self.assertEqual(len(tr), 41)

        # try a series of regex patterns that should return the same data
        st = client.get_waveforms_nscl("IUANMO BH", t, duration)
        patterns = ["IUANMO BH...",
                    "IUANMO BH.*",
                    "IUANMO BH[Z12].*",
                    "IUANMO BH[Z12]..",
                    "..ANMO BH.*"]
        for pattern in patterns:
            st2 = client.get_waveforms_nscl(pattern, t, duration)
            self.assertEqual(st, st2)


def suite():
    return unittest.makeSuite(ClientTestCase, 'test')


if __name__ == '__main__':
    unittest.main(defaultTest='suite')



"""
obspy.io.pde - NEIC PDE Bulletin read support for ObsPy
=======================================================

This module provides read support for NEIC Preliminary Determination of
Epicenters (PDE) Bulletin.

Currently, only the mchedr (machine readable Earthquake Data Report)
format is supported.

.. seealso:: http://earthquake.usgs.gov/data/pde.php

:copyright:
    The ObsPy Development Team (devs@obspy.org), Claudio Satriano
:license:
    GNU Lesser General Public License, Version 3
    (https://www.gnu.org/copyleft/lesser.html)

"""
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)
from future.builtins import *  # NOQA


if __name__ == '__main__':
    import doctest
    doctest.testmod(exclude_empty=True)



# -*- coding: utf-8 -*-
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)
from future.builtins import *  # NOQA

import unittest
from copy import deepcopy

import numpy as np

from obspy import Trace, UTCDateTime, read
from obspy.signal.filter import (bandpass, bandstop, highpass, lowpass,
                                 lowpass_cheby_2)
from obspy.signal.invsim import simulate_seismometer


class TraceTestCase(unittest.TestCase):
    """
    Test suite for obspy.core.trace.Trace.
    """
    def test_simulate(self):
        """
        Tests if calling simulate of trace gives the same result as using
        simulate_seismometer manually.
        """
        tr = read()[0]
        paz_sts2 = {'poles': [-0.037004 + 0.037016j, -0.037004 - 0.037016j,
                              - 251.33 + 0j, -131.04 - 467.29j,
                              - 131.04 + 467.29j],
                    'zeros': [0j, 0j],
                    'gain': 60077000.0,
                    'sensitivity': 2516778400.0}
        paz_le3d1s = {'poles': [-4.440 + 4.440j, -4.440 - 4.440j,
                                - 1.083 + 0.0j],
                      'zeros': [0.0 + 0.0j, 0.0 + 0.0j, 0.0 + 0.0j],
                      'gain': 0.4,
                      'sensitivity': 1.0}
        data = simulate_seismometer(
            tr.data, tr.stats.sampling_rate, paz_remove=paz_sts2,
            paz_simulate=paz_le3d1s, remove_sensitivity=True,
            simulate_sensitivity=True)
        tr.simulate(paz_remove=paz_sts2, paz_simulate=paz_le3d1s)
        # There is some strange issue on Win32bit (see #2188). Thus we just
        # use assert_allclose() here instead of testing for full equality.
        np.testing.assert_allclose(tr.data, data)

    def test_filter(self):
        """
        Tests the filter method of the Trace object.

        Basically three scenarios are tested (with differing filter options):
        - filtering with in_place=False:
            - is original trace unchanged?
            - is data of filtered trace the same as if done by hand
            - is processing information present in filtered trace
        - filtering with in_place=True:
            - is data of filtered trace the same as if done by hand
            - is processing information present in filtered trace
        - filtering with bad arguments passed to trace.filter():
            - is a TypeError properly raised?
            - after all bad filter calls, is the trace still unchanged?
        """
        # create two test Traces
        traces = []
        np.random.seed(815)
        header = {'network': 'BW', 'station': 'BGLD',
                  'starttime': UTCDateTime(2007, 12, 31, 23, 59, 59, 915000),
                  'npts': 412, 'sampling_rate': 200.0,
                  'channel': 'EHE'}
        traces.append(Trace(data=np.random.randint(0, 1000, 412),
                            header=deepcopy(header)))
        header['starttime'] = UTCDateTime(2008, 1, 1, 0, 0, 4, 35000)
        header['npts'] = 824
        traces.append(Trace(data=np.random.randint(0, 1000, 824),
                            header=deepcopy(header)))
        traces_bkp = deepcopy(traces)
        # different sets of filters to run test on:
        filters = [['bandpass', {'freqmin': 1., 'freqmax': 20.}],
                   ['bandstop', {'freqmin': 5, 'freqmax': 15., 'corners': 6}],
                   ['lowpass', {'freq': 30.5, 'zerophase': True}],
                   ['highpass', {'freq': 2, 'corners': 2}]]
        filter_map = {'bandpass': bandpass, 'bandstop': bandstop,
                      'lowpass': lowpass, 'highpass': highpass}

        # tests for in_place=True
        for i, tr in enumerate(traces):
            for filt_type, filt_ops in filters:
                tr = deepcopy(traces_bkp[i])
                tr.filter(filt_type, **filt_ops)
                # test if trace was filtered as expected
                data_filt = filter_map[filt_type](
                    traces_bkp[i].data,
                    df=traces_bkp[i].stats.sampling_rate, **filt_ops)
                np.testing.assert_array_equal(tr.data, data_filt)
                self.assertIn('processing', tr.stats)
                self.assertEqual(len(tr.stats.processing), 1)
                self.assertIn("filter", tr.stats.processing[0])
                self.assertIn(filt_type, tr.stats.processing[0])
                for key, value in filt_ops.items():
                    self.assertTrue("'%s': %s" % (key, value)
                                    in tr.stats.processing[0])
                # another filter run
                tr.filter(filt_type, **filt_ops)
                data_filt = filter_map[filt_type](
                    data_filt,
                    df=traces_bkp[i].stats.sampling_rate, **filt_ops)
                np.testing.assert_array_equal(tr.data, data_filt)
                self.assertIn('processing', tr.stats)
                self.assertEqual(len(tr.stats.processing), 2)
                for proc_info in tr.stats.processing:
                    self.assertIn("filter", proc_info)
                    self.assertIn(filt_type, proc_info)
                    for key, value in filt_ops.items():
                        self.assertTrue("'%s': %s" % (key, value)
                                        in proc_info)

        # some tests that should raise an Exception
        tr = traces[0]
        bad_filters = [
            ['bandpass', {'freqmin': 1., 'XXX': 20.}],
            ['bandstop', {'freqmin': 5, 'freqmax': "XXX", 'corners': 6}],
            ['bandstop', {}],
            ['bandstop', [1, 2, 3, 4, 5]],
            ['bandstop', None],
            ['bandstop', 3],
            ['bandstop', 'XXX'],
            ['bandpass', {'freqmin': 5, 'corners': 6}],
            ['bandpass', {'freqmin': 5, 'freqmax': 20., 'df': 100.}]]
        for filt_type, filt_ops in bad_filters:
            self.assertRaises(TypeError, tr.filter, filt_type, filt_ops)
        bad_filters = [['XXX', {'freqmin': 5, 'freqmax': 20., 'corners': 6}]]
        for filt_type, filt_ops in bad_filters:
            self.assertRaises(ValueError, tr.filter, filt_type, **filt_ops)
        # test if trace is unchanged after all these bad tests
        np.testing.assert_array_equal(tr.data, traces_bkp[0].data)
        self.assertEqual(tr.stats, traces_bkp[0].stats)

    def test_decimate(self):
        """
        Tests the decimate method of the Trace object.
        """
        # create test Trace
        tr = Trace(data=np.arange(20))
        tr_bkp = deepcopy(tr)
        # some test that should fail and leave the original trace alone
        self.assertRaises(ValueError, tr.decimate, 7, strict_length=True)
        self.assertRaises(ValueError, tr.decimate, 9, strict_length=True)
        self.assertRaises(ArithmeticError, tr.decimate, 18)
        # some tests in place
        tr.decimate(4, no_filter=True)
        np.testing.assert_array_equal(tr.data, np.arange(0, 20, 4))
        self.assertEqual(tr.stats.npts, 5)
        self.assertEqual(tr.stats.sampling_rate, 0.25)
        self.assertIn("decimate", tr.stats.processing[0])
        self.assertIn("factor=4", tr.stats.processing[0])
        tr = tr_bkp.copy()
        tr.decimate(10, no_filter=True)
        np.testing.assert_array_equal(tr.data, np.arange(0, 20, 10))
        self.assertEqual(tr.stats.npts, 2)
        self.assertEqual(tr.stats.sampling_rate, 0.1)
        self.assertIn("decimate", tr.stats.processing[0])
        self.assertIn("factor=10", tr.stats.processing[0])
        # some tests with automatic prefiltering
        tr = tr_bkp.copy()
        tr2 = tr_bkp.copy()
        tr.decimate(4)
        df = tr2.stats.sampling_rate
        tr2.data, fp = lowpass_cheby_2(data=tr2.data, freq=df * 0.5 / 4.0,
                                       df=df, maxorder=12, ba=False,
                                       freq_passband=True)
        # check that iteratively determined pass band frequency is correct
        self.assertAlmostEqual(0.0811378285461, fp, places=7)
        tr2.decimate(4, no_filter=True)
        np.testing.assert_array_equal(tr.data, tr2.data)


def suite():
    return unittest.makeSuite(TraceTestCase, 'test')


if __name__ == '__main__':
    unittest.main(defaultTest='suite')



# -*- coding: utf-8 -*-
"""
obspy.clients.syngine - IRIS Syngine client for ObsPy
=====================================================

This module offers methods to download from the IRIS syngine service
(https://ds.iris.edu/ds/products/syngine/). The service is able to generate
fully three dimensional synthetics through various 1D Earth models with
arbitrary source-receiver geometries and source mechanisms.

:copyright:
    The ObsPy Development Team (devs@obspy.org)
:license:
    GNU Lesser General Public License, Version 3
    (https://www.gnu.org/copyleft/lesser.html)


Basic Usage
-----------

First initialize a client object.

>>> from obspy.clients.syngine import Client
>>> client = Client()

Then request some data.

>>> st = client.get_waveforms(model="ak135f_5s", network="IU", station="ANMO",
...                           eventid="GCMT:C201002270634A")
>>> print(st)  # doctest: +ELLIPSIS
3 Trace(s) in Stream:
IU.ANMO.SE.MXZ | 2010-02-27T06:35:14... - ... | 4.0 Hz, 15520 samples
IU.ANMO.SE.MXN | 2010-02-27T06:35:14... - ... | 4.0 Hz, 15520 samples
IU.ANMO.SE.MXE | 2010-02-27T06:35:14... - ... | 4.0 Hz, 15520 samples
>>> st.plot()  # doctest: +SKIP

.. plot::

    from obspy.clients.syngine import Client
    client = Client()
    st = client.get_waveforms(model="ak135f_5s", network="IU", station="ANMO",
                              eventid="GCMT:C201002270634A")
    st.plot()

The available parameters are explained in detail in the
:meth:`~obspy.clients.syngine.client.Client.get_waveforms()` method and on
the `Syngine <https://ds.iris.edu/ds/products/syngine/>`_ website.


The queries are quite flexible. The following uses a station name wildcard
and only requests data around the P arrival. Please be a bit careful as one
can potentially download a lot of data with a single request.

>>> st = client.get_waveforms(model="ak135f_5s", network="IU", station="AN*",
...                           eventid="GCMT:C201002270634A",
...                           starttime="P-10", endtime="P+20")
>>> st.plot()  # doctest: +SKIP

.. plot::

    from obspy.clients.syngine import Client
    client = Client()
    st = client.get_waveforms(model="ak135f_5s", network="IU", station="A*",
                              eventid="GCMT:C201002270634A",
                              starttime="P-10", endtime="P+20")
    st.plot()


Bulk Requests
-------------

It is also possible to send requests for multiple user-defined stations. See
the :meth:`~obspy.clients.syngine.client.Client.get_waveforms_bulk()` method
for details. This example specifies a bunch of receiver coordinates and
requests seismograms for all of these.

>>> bulk = [{"latitude": 10.1, "longitude": 12.2, "stationcode": "AA"},
...         {"latitude": 14.5, "longitude": 10.0, "stationcode": "BB"}]
>>> st = client.get_waveforms_bulk(
...     model="ak135f_5s", eventid="GCMT:C201002270634A",
...     bulk=bulk, starttime="P-10", endtime="P+20")
>>> print(st)  # doctest: +ELLIPSIS
6 Trace(s) in Stream:
XX.AA.SE.MXZ | 2010-02-27T06:48:11... - ... | 4.0 Hz, 120 samples
XX.AA.SE.MXN | 2010-02-27T06:48:11... - ... | 4.0 Hz, 120 samples
XX.AA.SE.MXE | 2010-02-27T06:48:11... - ... | 4.0 Hz, 120 samples
XX.BB.SE.MXZ | 2010-02-27T06:48:15... - ... | 4.0 Hz, 120 samples
XX.BB.SE.MXN | 2010-02-27T06:48:15... - ... | 4.0 Hz, 120 samples
XX.BB.SE.MXE | 2010-02-27T06:48:15... - ... | 4.0 Hz, 120 samples


Other Useful Methods
--------------------

Use the :meth:`~obspy.clients.syngine.client.Client.get_available_models()`
method for a list of all available methods including some meta-information.

>>> client.get_available_models()  # doctest: +SKIP
{'ak135f_1s': {'components': 'vertical only',
  'default_components': 'Z',
  'default_dt': '0.05',
  'description': 'ak135 with density & Q of Montagner & Kennet(1996)',
  'length': 1815.00690721715,
  'max_event_depth': 750000,
  'max_period': '~100',
  'max_sampling_period': '0.255383',
  'min_period': 1.04999995231628},
 'ak135f_2s': ...
 ...
 }

The :func:`~obspy.clients.syngine.client.Client.get_model_info()`
method should be used to retrieve information about a specific model.

>>> from obspy.clients.syngine import Client
>>> c = Client()
>>> db_info = c.get_model_info(model_name="ak135f_5s")
>>> print(db_info.period)
5.125
"""
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)
from future.builtins import *  # NOQA
from future.utils import native_str

from .client import Client  # NOQA

__all__ = [native_str("Client")]


if __name__ == '__main__':
    import doctest
    doctest.testmod(exclude_empty=True)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Utility functions required for the download helpers.

:copyright:
    Lion Krischer (krischer@geophysik.uni-muenchen.de), 2014-2105
:license:
    GNU Lesser General Public License, Version 3
    (https://www.gnu.org/copyleft/lesser.html)
"""
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)
from future.builtins import *  # NOQA

import collections
import fnmatch
import itertools
import os
import sys
from lxml import etree
import numpy as np
from scipy.spatial import cKDTree
from socket import timeout as socket_timeout

if sys.version_info.major == 2:
    from urllib2 import HTTPError, URLError
else:
    from urllib.error import HTTPError, URLError

import obspy
from obspy.core import compatibility
from obspy.core.util.base import NamedTemporaryFile
from obspy.clients.fdsn.client import FDSNException
from obspy.io.mseed.util import get_record_information


# Different types of errors that can happen when downloading data via the
# FDSN clients.
ERRORS = [FDSNException, HTTPError, URLError, socket_timeout]

# Python 2 does have special classes for connection errors.
if sys.version_info.major == 2:
    # Use the base OS error on Python 2.
    ERRORS.append(OSError)
else:
    # All other connection errors inherit from it.
    ERRORS.append(ConnectionError)  # NOQA

ERRORS = tuple(ERRORS)

# mean earth radius in meter as defined by the International Union of
# Geodesy and Geophysics. Used for the spherical kd-tree.
EARTH_RADIUS = 6371009


ChannelAvailability = collections.namedtuple(
    "ChannelAvailability",
    ["network", "station", "location", "channel", "starttime", "endtime",
     "filename"])


def download_stationxml(client, client_name, bulk, filename, logger):
    """
    Download all channels for a station in the already prepared bulk list.

    :param client: An active client instance.
    :param client_name: The name of the client mainly used for logging
        purposes.
    :param bulk: An already prepared bulk download list for all channels and
        time intervals for the given station. All items in there are assumed
        to come from the same station.
    :param filename: The filename to download to.
    :param logger: The logger instance to use for logging.

    :returns: A tuple with the network and station id and the filename upon
        success
    """
    network = bulk[0][0]
    station = bulk[0][1]
    try:
        client.get_stations_bulk(bulk=bulk, level="response",
                                 filename=filename)
    except Exception:
        logger.info("Failed to download StationXML from '%s' for station "
                    "'%s.%s'." % (client_name, network, station))
        return None
    logger.info("Client '%s' - Successfully downloaded '%s'." %
                (client_name, filename))
    return ((network, station), filename)


def download_and_split_mseed_bulk(client, client_name, chunks, logger):
    """
    Downloads the channels of a list of stations in bulk, saves it to a
    temporary folder and splits it at the record level to obtain the final
    MiniSEED files.

    The big advantage of this approach is that it does not mess with the
    MiniSEED files at all. Each record, including all blockettes, will end
    up in the final files as they are served from the data centers.

    :param client: An active client instance.
    :param client_name: The name of the client instance used for logging
        purposes.
    :param chunks: A list of tuples, each denoting a single MiniSEED chunk.
        Each chunk is a tuple of network, station, location, channel,
        starttime, endtime, and desired filename.
    :param logger: An active logger instance.
    """
    # Create a dictionary of channel ids, each containing a list of
    # intervals, each of which will end up in a separate file.
    filenames = collections.defaultdict(list)
    for chunk in chunks:
        candidate = {
            "starttime": chunk[4],
            "endtime": chunk[5],
            "filename": chunk[6],
            "current_latest_endtime": None,
            "sequence_number": None}
        # Should not be necessary if chunks have been deduplicated before but
        # better safe than sorry.
        if candidate in filenames[tuple(chunk[:4])]:
            continue
        filenames[tuple(chunk[:4])].append(candidate)

    sequence_number = [0]

    def get_filename(starttime, endtime, c):
        """
        Helper function finding the corresponding filename in all filenames.

        :param starttime: The start time of the record.
        :param endtime: The end time of the record.
        :param c: A list of candidates.
        """
        # Make two passes. First find all candidates. This assumes that a
        # record cannot be larger than a single desired time interval. This
        # is probably always given except if somebody wants to download
        # files split into 1 second intervals...
        candidates = [
            _i for _i in c if
            (_i["starttime"] <= starttime <= _i["endtime"]) or
            (_i["starttime"] <= endtime <= _i["endtime"])]
        if not candidates:
            return None

        # If more then one candidate, apply some heuristics to find the
        # correct time interval. The main complication arises when the same
        # record is downloaded twice as it overlaps into two different
        # requested time intervals.
        if len(candidates) == 2:
            candidates.sort(key=lambda x: x["starttime"])
            first, second = candidates

            # Make sure the assumptions about the type of overlap are correct.
            if starttime > first["endtime"] or endtime < second["starttime"]:
                raise NotImplementedError

            # It must either be the last record of the first, or the first
            # record of the second candidate.
            if first["sequence_number"] is None and \
                    second["sequence_number"] is None:
                candidates = [second]

            # Unlikely to happen. Only if nothing but the very last record
            # of the first interval was available and the second interval
            # was first in the file.
            elif first["sequence_number"] is None:
                candidates = [first]

            # This is fairly likely and requires an additional check with
            # the latest time in the first interval.
            elif second["sequence_number"] is None:
                if starttime <= first["current_latest_endtime"]:
                    candidates = [second]
                else:
                    candidates = [first]

            # Neither are None. Just use the one with the higher sequence
            # number. This probably does not happen. If it happens something
            # else is a bit strange.
            else:
                if first["sequence_number"] > second["sequence_number"]:
                    candidates = [first]
                else:
                    candidates = [second]
        elif len(candidates) >= 2:
            raise NotImplementedError(
                "Please contact the developers. candidates: %s" %
                str(candidates))

        # Finally found the correct chunk
        ret_val = candidates[0]

        # Increment sequence number and make sure the current chunk is aware
        # of it.
        sequence_number[0] += 1
        ret_val["sequence_number"] = sequence_number[0]

        # Also write the time of the last chunk to it if necessary.
        ce = ret_val["current_latest_endtime"]
        if not ce or endtime > ce:
            ret_val["current_latest_endtime"] = endtime

        return ret_val["filename"]

    # Only the filename is not needed for the actual data request.
    bulk = [list(_i[:-1]) for _i in chunks]
    original_bulk_length = len(bulk)

    # Merge adjacent bulk-request for continuous downloads. This is a bit
    # redundant after splitting it up before, but eases the logic in the
    # other parts and puts less strain on the data centers' FDSN
    # implementation. It furthermore avoid the repeated download of records
    # that are part of two neighbouring time intervals.
    bulk_channels = collections.defaultdict(list)
    for b in bulk:
        bulk_channels[(b[0], b[1], b[2], b[3])].append(b)

    # Merge them.
    for key, value in bulk_channels.items():
        # Sort based on starttime.
        value = sorted(value, key=lambda x: x[4])
        # Merge adjacent.
        cur_bulk = value[0:1]
        for b in value[1:]:
            # Random threshold of 2 seconds. Reasonable for most real world
            # cases.
            if b[4] <= cur_bulk[-1][5] + 2:
                cur_bulk[-1][5] = b[5]
                continue
            cur_bulk.append(b)
        bulk_channels[key] = cur_bulk
    bulk = list(itertools.chain.from_iterable(bulk_channels.values()))

    # Save first to a temporary file, then cut the file into separate files.
    with NamedTemporaryFile() as tf:
        temp_filename = tf.name
        open_files = {}

        client.get_waveforms_bulk(bulk, filename=temp_filename)
        # If that succeeds, split the old file into multiple new ones.
        file_size = os.path.getsize(temp_filename)

        with open(temp_filename, "rb") as fh:
            try:
                while True:
                    if fh.tell() >= (file_size - 256):
                        break
                    info = get_record_information(fh)
                    channel_id = (info["network"], info["station"],
                                  info["location"], info["channel"])

                    # Sometimes the services return something nobody wants...
                    if channel_id not in filenames:
                        fh.read(info["record_length"])
                        continue
                    # Get the best matching filename.
                    filename = get_filename(
                        starttime=info["starttime"], endtime=info["endtime"],
                        c=filenames[channel_id])
                    # Again sometimes there are time ranges nobody asked for...
                    if filename is None:
                        fh.read(info["record_length"])
                        continue
                    if filename not in open_files:
                        open_files[filename] = open(filename, "wb")
                    open_files[filename].write(fh.read(info["record_length"]))
            finally:
                for f in open_files.values():
                    try:
                        f.close()
                    except Exception:
                        pass
    logger.info("Client '%s' - Successfully downloaded %i channels (of %i)" % (
        client_name, len(open_files), original_bulk_length))
    return sorted(open_files.keys())


class SphericalNearestNeighbour(object):
    """
    Spherical nearest neighbour queries using scipy's fast kd-tree
    implementation.
    """
    def __init__(self, data):
        cart_data = self.spherical2cartesian(data)
        self.data = data
        self.kd_tree = cKDTree(data=cart_data, leafsize=10)

    def query(self, points):
        points = self.spherical2cartesian(points)
        d, i = self.kd_tree.query(points)

        # Filter NaNs. Happens when not enough points are available.
        m = np.isfinite(d)
        return d[m], i[m]

    def query_pairs(self, maximum_distance):
        return self.kd_tree.query_pairs(maximum_distance)

    @staticmethod
    def spherical2cartesian(data):
        """
        Converts a list of :class:`~obspy.fdsn.download_status.Station`
        objects to an array of shape(len(list), 3) containing x/y/z in meters.
        """
        # Create three arrays containing lat/lng/radius.
        shape = len(data)
        lat = np.array([_i.latitude for _i in data], dtype=np.float64)
        lon = np.array([_i.longitude for _i in data], dtype=np.float64)
        r = np.array([EARTH_RADIUS for _ in data], dtype=np.float64)
        # Convert data from lat/lng to x/y/z.
        colat = 90.0 - lat
        cart_data = np.empty((shape, 3), dtype=np.float64)

        cart_data[:, 0] = r * np.sin(np.deg2rad(colat)) * \
            np.cos(np.deg2rad(lon))
        cart_data[:, 1] = r * np.sin(np.deg2rad(colat)) * \
            np.sin(np.deg2rad(lon))
        cart_data[:, 2] = r * np.cos(np.deg2rad(colat))

        return cart_data


def filter_channel_priority(channels, key, priorities=None):
    """
    This function takes a dictionary containing channels keys and returns a new
    one filtered with the given priorities list.

    All channels matching the first pattern in the list will be retrieved. If
    one or more channels are found it stops. Otherwise it will attempt to
    retrieve channels matching the next pattern. And so on.

    :type channels: list
    :param channels: A list containing channel names.
    :type priorities: list of unicode or None
    :param priorities: The desired channels with descending priority. Channels
    will be matched by fnmatch.fnmatch() so wildcards and sequences are
    supported. The advisable form to request the three standard components
    of a channel is "HH[ZNE]" to avoid getting e.g. rotated components.
    :returns: A new list containing only the filtered channels.
    """
    if priorities is None:
        return channels
    filtered_channels = []
    for pattern in priorities:
        if filtered_channels:
            break
        for channel in channels:
            if fnmatch.fnmatch(getattr(channel, key), pattern):
                filtered_channels.append(channel)
                continue
    return filtered_channels


def safe_delete(filename):
    """
    "Safely" delete a file. It really just checks if it exists and if it is a
    file.

    :param filename: The filename to delete.
    """
    if not os.path.exists(filename):
        return
    elif not os.path.isfile(filename):
        raise ValueError("'%s' is not a file." % filename)
    try:
        os.remove(filename)
    except Exception as e:
        raise ValueError("Could not delete '%s' because: %s" % (filename,
                                                                str(e)))


def get_stationxml_contents(filename):
    """
    Really fast way to get all channels with a response in a StationXML
    file. Sometimes it is too expensive to parse the full file with ObsPy.
    This is usually orders of magnitudes faster.

    Will only returns channels that contain response information.

    :param filename: The path to the file.
    :returns: list of ChannelAvailability objects.
    """
    # Small state machine.
    network, station, location, channel, starttime, endtime = [None] * 6

    ns = "http://www.fdsn.org/xml/station/1"
    network_tag = "{%s}Network" % ns
    station_tag = "{%s}Station" % ns
    channel_tag = "{%s}Channel" % ns
    response_tag = "{%s}Response" % ns

    try:
        context = etree.iterparse(filename, events=("start", ),
                                  tag=(network_tag, station_tag, channel_tag,
                                       response_tag))
    except TypeError:  # pragma: no cover
        # Some old lxml version have a way less powerful iterparse()
        # function. Fall back to parsing with ObsPy.
        return _get_stationxml_contents_slow(filename)

    channels = []
    for event, elem in context:
        if elem.tag == channel_tag:
            channel = elem.get('code')
            location = elem.get('locationCode').strip()
            starttime = obspy.UTCDateTime(elem.get('startDate'))
            end_date = elem.get('endDate')
            if end_date is not None:
                endtime = obspy.UTCDateTime(elem.get('endDate'))
            else:
                endtime = obspy.UTCDateTime(2599, 1, 1)
        elif elem.tag == response_tag:
            channels.append(ChannelAvailability(
                network, station, location, channel, starttime, endtime,
                filename))
        elif elem.tag == station_tag:
            station = elem.get('code')
            location, channel, starttime, endtime = \
                None, None, None, None
        elif elem.tag == network_tag:
            network = elem.get('code')
            station, location, channel, starttime, endtime = \
                None, None, None, None, None
        while elem.getprevious() is not None:
            del elem.getparent()[0]

    return channels


def _get_stationxml_contents_slow(filename):
    """
    Slow variant for get_stationxml_contents() for old lxml versions.

    :param filename: The path to the file.
    :returns: list of ChannelAvailability objects.
    """
    channels = []
    inv = obspy.read_inventory(filename)
    for net in inv:
        for sta in net:
            for cha in sta:
                if not cha.response:
                    continue
                channels.append(ChannelAvailability(
                    net.code, sta.code, cha.location_code, cha.code,
                    cha.start_date, cha.end_date
                    if cha.end_date else obspy.UTCDateTime(2599, 1, 1),
                    filename))
    return channels


def get_stationxml_filename(str_or_fct, network, station, channels,
                            starttime, endtime):
    """
    Helper function getting the filename of a StationXML file.

    :param str_or_fct: The string or function to be evaluated.
    :type str_or_fct: function or str
    :param network: The network code.
    :type network: str
    :param station: The station code.
    :type station: str
    :param channels: The channels. Each channel is a tuple of two strings:
        location code and channel code.
    :type channels: list of tuples
    :param starttime: The start time.
    :type starttime: :class:`~obspy.core.utcdatetime.UTCDateTime`
    :param endtime: The end time.
    :type endtime: :class:`~obspy.core.utcdatetime.UTCDateTime`

    The rules are simple, if it is a function, then the network, station,
    channels, start time, and end time parameters are passed to it.

    If it is a string, and it contains ``"{network}"``, and ``"{station}"``
    formatting specifiers, ``str.format()`` is called.

    Otherwise it is considered to be a folder name and the resulting
    filename will be ``"FOLDER_NAME/NET.STA.xml"``

    This function will either return a string or a dictionary with three
    keys: ``"available_channels"``, ``"missing_channels"``, and ``"filename"``.
    """
    # Call if possible.
    if callable(str_or_fct):
        path = str_or_fct(network, station, channels, starttime, endtime)
    # Check if it's a format template.
    elif ("{network}" in str_or_fct) and ("{station}" in str_or_fct):
        path = str_or_fct.format(network=network, station=station)
    # Otherwise assume it's a path.
    else:
        path = os.path.join(str_or_fct, "{network}.{station}.xml".format(
            network=network, station=station))

    # If it is just a filename, return that.
    if isinstance(path, (str, bytes)):
        return path

    elif isinstance(path, compatibility.collections_abc.Container):
        if "available_channels" not in path or \
                "missing_channels" not in path or \
                "filename" not in path:
            raise ValueError(
                "The dictionary returned by the stationxml filename function "
                "must contain the following keys: 'available_channels', "
                "'missing_channels', and 'filename'.")
        if not isinstance(path["available_channels"],
                          compatibility.collections_abc.Iterable) or\
                not isinstance(path["missing_channels"],
                               compatibility.collections_abc.Iterable) or \
                not isinstance(path["filename"], (str, bytes)):
            raise ValueError("Return types must be two lists of channels and "
                             "a string for the filename.")
        return path

    else:
        raise TypeError("'%s' is not a filepath." % str(path))


def get_mseed_filename(str_or_fct, network, station, location, channel,
                       starttime, endtime):
    """
    Helper function getting the filename of a MiniSEED file.

    The rules are simple, if it is a function, then the network, station,
    location, channel, start time, and end time parameters are passed to it
    and the resulting string is returned. If the return values is ``True``,
    the particular time interval will be ignored.

    If it is a string, and it contains ``"{network}"``,  ``"{station}"``,
    ``"{location}"``, ``"{channel}"``, ``"{starttime}"``, and ``"{endtime}"``
    formatting specifiers, ``str.format()`` is called.

    Otherwise it is considered to be a folder name and the resulting
    filename will be
    ``"FOLDER_NAME/NET.STA.LOC.CHAN__STARTTIME__ENDTIME.mseed"``

    In the last two cases, the times will be formatted with
    ``"%Y%m%dT%H%M%SZ"``.
    """
    strftime = "%Y%m%dT%H%M%SZ"
    if callable(str_or_fct):
        path = str_or_fct(network, station, location, channel, starttime,
                          endtime)
    elif ("{network}" in str_or_fct) and ("{station}" in str_or_fct) and \
            ("{location}" in str_or_fct) and ("{channel}" in str_or_fct) and \
            ("{starttime}" in str_or_fct) and ("{endtime}" in str_or_fct):
        path = str_or_fct.format(
            network=network, station=station, location=location,
            channel=channel, starttime=starttime.strftime(strftime),
            endtime=endtime.strftime(strftime))
    else:
        path = os.path.join(
            str_or_fct,
            "{network}.{station}.{location}.{channel}__{s}__{e}.mseed".format(
                network=network, station=station, location=location,
                channel=channel, s=starttime.strftime(strftime),
                e=endtime.strftime(strftime)))

    if path is True:
        return True
    elif not isinstance(path, (str, bytes)):
        raise TypeError("'%s' is not a filepath." % str(path))
    return path


if __name__ == '__main__':
    import doctest
    doctest.testmod(exclude_empty=True)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
A command-line program that indexes seismogram files into a database.

:copyright:
    The ObsPy Development Team (devs@obspy.org)
:license:
    GNU Lesser General Public License, Version 3
    (https://www.gnu.org/copyleft/lesser.html)

.. rubric:: Usage Examples

(1) Run indexer as daemon continuously crawling the given paths but index only
    the last 24 hours (-r24) of a waveform archive::

       #!/bin/bash
       DB=postgresql://username:password@localhost:5432/database
       DATA=/path/to/archive/2010,/path/to/archive/2011,/path/to/arclink
       LOG=/path/to/indexer.log
       ./obspy-indexer -v -i0.0 -n1 -u$DB -d$DATA -r24 -l$LOG &

(2) Run only once and remove duplicates::

       ./obspy-indexer -v -i0.0 --run-once --check-duplicates -n1 -u$DB -d$DATA
"""
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)
from future.builtins import *  # NOQA @UnusedWildImport
from future.utils import native_str

import logging
import multiprocessing
import select
import sys
from argparse import ArgumentParser

if sys.version_info.major == 2:
    import BaseHTTPServer as http_server  # NOQA
else:
    import http.server as http_server

from sqlalchemy import create_engine
from sqlalchemy.orm.session import sessionmaker

from obspy import __version__
from obspy.db.db import Base
from obspy.db.indexer import WaveformFileCrawler, worker
from obspy.db.util import parse_mapping_data


class MyHandler(http_server.BaseHTTPRequestHandler):
    def do_GET(self):  # noqa
        """
        Respond to a GET request.
        """
        out = """<html>
  <head>
    <title>obspy-indexer status</title>
    <meta http-equiv="refresh" content="10" />
    <style type="text/css">
      th { text-align: left; font-family:monospace; width: 150px;
           vertical-align: top; padding: 3px; }
      td { font-family:monospace; padding: 3px;}
      pre { margin: 0; }
    </style>
  </head>
  <body>
    <h1>obspy-indexer</h1>
    <h2>Options</h2>
"""
        out += '<table>'
        for key, value in sorted(self.server.options.__dict__.items()):
            out += "<tr><th>%s</th><td>%s</td></tr>" % (key, value)
        if self.server.mappings:
            out += "<tr><th>mapping rules</th><td>%s</td></tr>" % \
                   (self.server.mappings)
        out += '</table>'
        out += '<h2>Status</h2>'
        out += '<table>'
        out += "<tr><th>current path</th><td>%s</td></tr>" % \
            (self.server._current_path)
        out += "<tr><th>patterns</th><td><pre>%s</pre></td></tr>" % \
            ('\n'.join(self.server.patterns))
        out += "<tr><th>features</th><td><pre>%s</pre></td></tr>" % \
            ('\n'.join(self.server.features))
        out += "<tr><th>file queue</th><td><pre>%s</pre></td></tr>" % \
            ('\n'.join(self.server._current_files))
        out += '</table>'
        out += "</body></html>"
        self.send_response(200)
        self.send_header("Content-type", "text/html")
        self.end_headers()
        self.wfile.write(out)


class WaveformIndexer(http_server.HTTPServer, WaveformFileCrawler):
    """
    A waveform indexer server.
    """

    def serve_forever(self, poll_interval=0.5):
        self.running = True
        while self.running:
            r, _w, _e = select.select([self], [], [], poll_interval)
            if r:
                self._handle_request_noblock()
            self.iterate()


def _run_indexer(options):
    logging.info("Starting indexer %s:%s ..." % (options.host, options.port))
    # initialize crawler
    service = WaveformIndexer((options.host, options.port), MyHandler)
    service.log = logging
    try:
        # prepare paths
        if ',' in options.data:
            paths = options.data.split(',')
        else:
            paths = [options.data]
        paths = service._prepare_paths(paths)
        if not paths:
            return
        # prepare map file
        if options.mapping_file:
            with open(options.mapping_file, 'r') as f:
                data = f.readlines()
            mappings = parse_mapping_data(data)
            logging.info("Parsed %d lines from mapping file %s" %
                         (len(data), options.mapping_file))
        else:
            mappings = {}
        # create file queue and worker processes
        manager = multiprocessing.Manager()
        in_queue = manager.dict()
        work_queue = manager.list()
        out_queue = manager.list()
        log_queue = manager.list()
        # spawn processes
        for i in range(options.number_of_cpus):
            args = (i, in_queue, work_queue, out_queue, log_queue, mappings)
            p = multiprocessing.Process(target=worker, args=args)
            p.daemon = True
            p.start()
        # connect to database
        engine = create_engine(options.db_uri, encoding=native_str('utf-8'),
                               convert_unicode=True)
        metadata = Base.metadata
        # recreate database
        if options.drop_database:
            metadata.drop_all(engine, checkfirst=True)
        metadata.create_all(engine, checkfirst=True)
        # initialize database + options
        _session = sessionmaker(bind=engine)
        service.session = _session
        service.options = options
        service.mappings = mappings
        # set queues
        service.input_queue = in_queue
        service.work_queue = work_queue
        service.output_queue = out_queue
        service.log_queue = log_queue
        service.paths = paths
        service._reset_walker()
        service._step_walker()
        service.serve_forever(options.poll_interval)
    except KeyboardInterrupt:
        quit()
    logging.info("Indexer stopped.")


def main(argv=None):
    parser = ArgumentParser(prog='obspy-indexer',
                            description='\n'.join(__doc__.split('\n')[:3]))
    parser.add_argument('-V', '--version', action='version',
                        version="%(prog)s " + __version__)
    parser.add_argument(
        '-v', '--verbose', action='store_true',
        help='Verbose output.')
    parser.add_argument(
        '-d', '--data', default='data=*.*',
        help="""Path, search patterns and feature plug-ins of waveform files.
The indexer will crawl recursively through all sub-directories within each
given path. Multiple paths have to be separated with a comma, e.g.
'/first/path=*.*,/second/path,/third/path=*.gse'.
File patterns are given as space-separated list of wildcards after a equal
sign, e.g.
'/path=*.gse2 *.mseed,/second/path=*.*'.
Feature plug-ins may be added using the hash (#) character, e.g.
'/path=*.mseed#feature1#feature2,/second/path#feature2'.
Be aware that features must be provided behind file patterns (if any)! There is
no default feature enabled.
Default path option is 'data=*.*'.""")
    parser.add_argument(
        '-u', '--db-uri', default='sqlite:///indexer.sqlite',
        help="Database connection URI, such as "
             "postgresql://scott:tiger@localhost/mydatabase. "
             "Default is a SQLite database './indexer.sqlite'.")
    parser.add_argument(
        '-n', type=int, dest='number_of_cpus',
        help="Number of CPUs used for the indexer.",
        default=multiprocessing.cpu_count())
    parser.add_argument(
        '-i', '--poll-interval', type=float, default=0.1,
        help="Poll interval for file crawler in seconds (default is 0.1).")
    parser.add_argument(
        '-r', '--recent', type=int, default=0,
        help="Index only recent files modified within the given "
             "number of hours. This option is deactivated by default.")
    parser.add_argument(
        '-l', '--log', default='',
        help="Log file name. If no log file is given, stdout will be used.")
    parser.add_argument(
        '-m', '--mapping-file', default=None,
        help="Correct network, station, location and channel codes using a" +
             " custom mapping file.")
    parser.add_argument(
        '-a', '--all-files', action='store_false', dest='skip_dots',
        help="The indexer will automatically skip paths or "
             "files starting with a dot. This option forces "
             "parsing of all paths and files.")
    parser.add_argument(
        '-1', '--run-once', action='store_true',
        help="The indexer will parse through all given directories only "
             "once and quit afterwards.")
    parser.add_argument(
        '--check-duplicates', action='store_true',
        help="Checks for duplicate entries within database. "
             "This feature will slow down the indexer progress.")
    parser.add_argument(
        '--cleanup', action='store_true',
        help="Clean database from non-existing files or paths " +
             "if activated, but will skip all paths marked as " +
             "archived in the database.")
    parser.add_argument(
        '-f', '--force-reindex', action='store_true',
        help="Reindex existing index entry for every crawled file.")
    parser.add_argument(
        '--drop-database', action='store_true',
        help="Deletes and recreates the complete database at start up.")
    parser.add_argument(
        '-H', '--host', default='localhost',
        help="Server host name. Default is 'localhost'.")
    parser.add_argument(
        '-p', '--port', type=int, default=0,
        help="Port number. If not given a free port will be picked.")

    args = parser.parse_args(argv)
    # set level of verbosity
    if args.verbose:
        level = logging.DEBUG
    else:
        level = logging.INFO
    if args.log == "":
        logging.basicConfig(stream=sys.stdout, level=level,
                            format="%(asctime)s [%(levelname)s] %(message)s")
    else:
        logging.basicConfig(filename=args.log, level=level,
                            format="%(asctime)s [%(levelname)s] %(message)s")
    _run_indexer(args)


if __name__ == "__main__":
    main()
